# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import functools

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
from oslo_service import service
from oslo_utils import timeutils
from oslo_utils import uuidutils
import six

from senlin.common import consts
from senlin.common import context as senlin_context
from senlin.common import exception
from senlin.common.i18n import _, _LE, _LI
from senlin.common import messaging as rpc_messaging
from senlin.common import scaleutils as su
from senlin.common import schema
from senlin.common import utils
from senlin.engine.actions import base as action_mod
from senlin.engine import cluster as cluster_mod
from senlin.engine import cluster_policy as cpm
from senlin.engine import dispatcher
from senlin.engine import environment
from senlin.engine import health_manager
from senlin.engine import node as node_mod
from senlin.engine.receivers import base as receiver_mod
from senlin.engine import scheduler
from senlin.objects import action as action_obj
from senlin.objects import cluster as cluster_obj
from senlin.objects import cluster_policy as cp_obj
from senlin.objects import credential as cred_obj
from senlin.objects import event as event_obj
from senlin.objects import node as node_obj
from senlin.objects import policy as policy_obj
from senlin.objects import profile as profile_obj
from senlin.objects import receiver as receiver_obj
from senlin.objects import service as service_obj
from senlin.policies import base as policy_base
from senlin.profiles import base as profile_base

LOG = logging.getLogger(__name__)

CONF = cfg.CONF


def request_context(func):
    @functools.wraps(func)
    def wrapped(self, ctx, *args, **kwargs):
        if ctx is not None and not isinstance(ctx,
                                              senlin_context.RequestContext):
            ctx = senlin_context.RequestContext.from_dict(ctx.to_dict())
        try:
            return func(self, ctx, *args, **kwargs)
        except exception.SenlinException:
            raise oslo_messaging.rpc.dispatcher.ExpectedException()
    return wrapped


class EngineService(service.Service):
    '''Lifecycle manager for a running service engine.

    - All the contained methods here are called from the RPC client.
    - If a RPC call does not have a corresponding method here, an exception
      will be thrown.
    - Arguments to these calls are added dynamically and will be treated as
      keyword arguments by the RPC client.
    '''

    def __init__(self, host, topic, manager=None):

        super(EngineService, self).__init__()
        self.host = host
        self.topic = topic
        self.dispatcher_topic = consts.ENGINE_DISPATCHER_TOPIC
        self.health_mgr_topic = consts.ENGINE_HEALTH_MGR_TOPIC

        # The following are initialized here and will be assigned in start()
        # which happens after the fork when spawning multiple worker processes
        self.engine_id = None
        self.TG = None
        self.target = None
        self._rpc_server = None

        # Intialize the global environment
        environment.initialize()

    def init_tgm(self):
        self.TG = scheduler.ThreadGroupManager()

    def start(self):
        self.engine_id = uuidutils.generate_uuid()
        self.init_tgm()

        # create a dispatcher RPC service for this engine.
        self.dispatcher = dispatcher.Dispatcher(self,
                                                self.dispatcher_topic,
                                                consts.RPC_API_VERSION,
                                                self.TG)
        LOG.info(_LI("Starting dispatcher for engine %s"), self.engine_id)
        self.dispatcher.start()

        target = oslo_messaging.Target(version=consts.RPC_API_VERSION,
                                       server=self.host,
                                       topic=self.topic)
        self.target = target
        self._rpc_server = rpc_messaging.get_rpc_server(target, self)
        self._rpc_server.start()
        self.service_manage_cleanup()

        # create a health manager RPC service for this engine.
        self.health_mgr = health_manager.HealthManager(
            self, self.health_mgr_topic, consts.RPC_API_VERSION)

        LOG.info(_LI("Starting health manager for engine %s"), self.engine_id)
        self.health_mgr.start()

        self.TG.add_timer(cfg.CONF.periodic_interval,
                          self.service_manage_report)
        super(EngineService, self).start()

    def _stop_rpc_server(self):
        # Stop RPC connection to prevent new requests
        LOG.info(_LI("Stopping engine service..."))
        try:
            self._rpc_server.stop()
            self._rpc_server.wait()
            LOG.info(_LI('Engine service stopped successfully'))
        except Exception as ex:
            LOG.error(_LE('Failed to stop engine service: %s'),
                      six.text_type(ex))

    def stop(self):
        self._stop_rpc_server()

        # Notify dispatcher to stop all action threads it started.
        LOG.info(_LI("Stopping dispatcher for engine %s"), self.engine_id)
        self.dispatcher.stop()

        # Notify health_manager to stop
        LOG.info(_LI("Stopping health manager for engine %s"), self.engine_id)
        self.health_mgr.stop()

        self.TG.stop()
        super(EngineService, self).stop()

    def service_manage_report(self):
        ctx = senlin_context.get_admin_context()
        try:
            svc = service_obj.Service.update(ctx, self.engine_id)
            # if svc is None, means it's not created.
            if svc is None:
                service_obj.Service.create(ctx, self.engine_id, self.host,
                                           'senlin-engine', self.topic)
        except Exception as ex:
            LOG.error(_LE('Service %(service_id)s update failed: %(error)s'),
                      {'service_id': self.engine_id, 'error': ex})

    def service_manage_cleanup(self):
        ctx = senlin_context.get_admin_context()
        time_window = (2 * cfg.CONF.periodic_interval)
        svcs = service_obj.Service.get_all(ctx)
        for svc in svcs:
            if svc['id'] == self.engine_id:
                continue
            if timeutils.is_older_than(svc['updated_at'], time_window):
                # < time_line:
                # hasn't been updated, assuming it's died.
                LOG.info(_LI('Service %s was aborted'), svc['id'])
                service_obj.Service.delete(ctx, svc['id'])

    @request_context
    def credential_create(self, context, cred, attrs=None):
        """Create the credential based on the context.

        We may add more parameters in future to the query parameter, for
        example as Senlin expands its support to non-OpenStack backends.

        :param context: The requesting context which contains the user id
                        along with other identity information.
        :param cred: A credential to be associated with the user identity
                     provided in the context.
        :param dict attrs: Optional attributes associated with the credential.
        :return: A dictionary containing the persistent credential.
        """
        values = {
            'user': context.user,
            'project': context.project,
            'cred': {
                'openstack': {
                    'trust': cred
                }
            }
        }
        cred_obj.Credential.update_or_create(context, values)
        return {'cred': cred}

    @request_context
    def credential_get(self, context, query=None):
        """Get the credential based on the context.

        We may add more parameters in future to the query parameter, for
        example as Senlin expands its support to non-OpenStack backends.

        :param context: The requesting context which contains the user id
            along with other identity information.
        :param dict query: Optional query parameters.
        :return: A dictionary containing the persistent credential, or None
            if no matching credential is found.
        """
        res = cred_obj.Credential.get(context, context.user, context.project)
        if res is None:
            return None
        return res.cred.get('openstack', None)

    @request_context
    def credential_update(self, context, cred, **attrs):
        """Update a credential based on the context and provided value.

        We may add more parameters in future to the query parameter, for
        example as Senlin expands its support to non-OpenStack backends.

        :param context: The requesting context which contains the user id
                        along with other identity information.
        :param dict attrs: Optional attribute values to be associated with
                           the credential.
        :return: A dictionary containing the updated credential.
        """
        cred_obj.Credential.update(context, context.user, context.project,
                                   {'cred': {'openstack': {'trust': cred}}})
        return {'cred': cred}

    @request_context
    def get_revision(self, context):
        return cfg.CONF.revision['senlin_engine_revision']

    @request_context
    def profile_type_list(self, context):
        """List known profile type implementations.

        :param context: An instance of the request context.
        :return: A list of profile types.
        """
        return environment.global_env().get_profile_types()

    @request_context
    def profile_type_get(self, context, type_name):
        """Get the details about a profile type.

        :param context: An instance of the request context.
        :param type_name: The name of a profile type.
        :return: The details about a profile type.
        """
        profile = environment.global_env().get_profile(type_name)
        data = profile.get_schema()

        return {
            'name': type_name,
            'schema': data,
        }

    def profile_find(self, context, identity, project_safe=True):
        """Find a profile with the given identity.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a profile.
        :param project_safe: A boolean indicating whether profile from
                             projects other than the requesting one can be
                             returned.
        :return: A DB object of profile or an exception `ResourceNotFound`
                 if no matching object is found.
        """
        if uuidutils.is_uuid_like(identity):
            profile = profile_obj.Profile.get(context, identity,
                                              project_safe=project_safe)
            if not profile:
                profile = profile_obj.Profile.get_by_name(
                    context, identity, project_safe=project_safe)
        else:
            profile = profile_obj.Profile.get_by_name(
                context, identity, project_safe=project_safe)
            if not profile:
                profile = profile_obj.Profile.get_by_short_id(
                    context, identity, project_safe=project_safe)

        if not profile:
            raise exception.ResourceNotFound(type='profile', id=identity)

        return profile

    @request_context
    def profile_list(self, context, limit=None, marker=None, sort=None,
                     filters=None, project_safe=True):
        """List profiles matching the specified criteria.

        :param context: An instance of request context.
        :param limit: An integer specifying the maximum number of profiles to
                      return in a response.
        :param marker: An UUID specifying the profile after which the result
                       list starts.
        :param sort: A list of sorting keys (each optionally attached with a
                     sorting direction) separated by commas.
        :param filters: A dictionary of key-value pairs for filtering out the
                        result list.
        :param project_safe: A boolean indicating whether profiles from all
                             projects will be returned.
        :return: A list of `Profile` object representations.
        """
        limit = utils.parse_int_param(consts.PARAM_LIMIT, limit)
        utils.validate_sort_param(sort, consts.PROFILE_SORT_KEYS)
        project_safe = utils.parse_bool_param('project_safe', project_safe)
        if not project_safe and not context.is_admin:
            raise exception.Forbidden()

        profiles = profile_base.Profile.load_all(context,
                                                 limit=limit, marker=marker,
                                                 sort=sort, filters=filters,
                                                 project_safe=project_safe)

        return [p.to_dict() for p in profiles]

    def _validate_profile(self, context, spec, name=None,
                          metadata=None, validate_props=False):
        """Validate a profile.

        :param context: An instance of the request context.
        :param name: The name for the profile to be created.
        :param spec: A dictionary containing the spec for the profile.
        :param metadata: A dictionary containing optional key-value pairs to
                         be associated with the profile.
        :param validate_props: Whether to validate if provide a valid Value
                               to property.
        :return: Validated profile object.
        """
        type_name, version = schema.get_spec_version(spec)
        type_str = "-".join([type_name, version])
        try:
            plugin = environment.global_env().get_profile(type_str)
        except exception.ResourceNotFound as ex:
            msg = ex.enhance_msg('specified', ex)
            raise exception.SpecValidationFailed(message=msg)

        kwargs = {
            'user': context.user,
            'project': context.project,
            'domain': context.domain,
            'metadata': metadata
        }
        if name is None:
            name = 'validated_profile'
        profile = plugin(name, spec, **kwargs)
        try:
            profile.validate(validate_props=validate_props)
        except exception.InvalidSpec as ex:
            msg = six.text_type(ex)
            LOG.error(_LE("Failed in validating profile: %s"), msg)
            raise exception.SpecValidationFailed(message=msg)

        return profile

    @request_context
    def profile_create(self, context, name, spec, metadata=None):
        """Create a profile with the given properties.

        :param context: An instance of the request context.
        :param name: The name for the profile to be created.
        :param spec: A dictionary containing the spec for the profile.
        :param metadata: A dictionary containing optional key-value pairs to
                         be associated with the profile.
        :return: A dictionary containing the details of the profile object
                 created.
        """
        if cfg.CONF.name_unique:
            if profile_obj.Profile.get_by_name(context, name):
                msg = _("A profile named '%(name)s' already exists."
                        ) % {"name": name}
                raise exception.BadRequest(msg=msg)

        profile = self._validate_profile(context, spec, name=name,
                                         metadata=metadata)

        LOG.info(_LI("Creating profile %(type)s '%(name)s'."),
                 {'type': profile.type, 'name': profile.name})

        profile.store(context)

        LOG.info(_LI("Profile %(name)s is created: %(id)s."),
                 {'name': name, 'id': profile.id})

        return profile.to_dict()

    @request_context
    def profile_validate(self, context, spec):
        """Validate a profile with the given properties.

        :param context: An instance of the request context.
        :param spec: A dictionary containing the spec for the profile.
        :return: A dictionary containing the details of the profile object
                 validated.
        """
        profile = self._validate_profile(context, spec, validate_props=True)

        return profile.to_dict()

    @request_context
    def profile_get(self, context, identity):
        """Retrieve the details about a profile.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a profile.
        :return: A dictionary containing the policy details, or an exception
                 of type `ResourceNotFound` if no matching object is found.
        """
        db_profile = self.profile_find(context, identity)
        profile = profile_base.Profile.load(context, profile=db_profile)
        return profile.to_dict()

    @request_context
    def profile_update(self, context, profile_id, name=None, metadata=None):
        """Update the properties of a given profile.

        :param context: An instance of the request context.
        :param profile_id: The UUID, name or short-id of a profile.
        :param name: The new name for the profile.
        :param metadata: A dictionary of key-value pairs to be associated with
                         the profile.
        :returns: A dictionary containing the details of the updated profile,
                  or an exception `ResourceNotFound` if no matching profile is
                  found.
        """
        LOG.info(_LI("Updating profile '%(id)s.'"), {'id': profile_id})

        db_profile = self.profile_find(context, profile_id)
        profile = profile_base.Profile.load(context, profile=db_profile)
        changed = False
        if name is not None and name != profile.name:
            profile.name = name
            changed = True
        if metadata is not None and metadata != profile.metadata:
            profile.metadata = metadata
            changed = True
        if changed:
            profile.store(context)
        else:
            msg = _("No property needs an update.")
            raise exception.BadRequest(msg=msg)

        LOG.info(_LI("Profile '%(id)s' is updated."), {'id': profile_id})
        return profile.to_dict()

    @request_context
    def profile_delete(self, context, identity):
        """Delete the specified profile.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a profile.
        :return: None if succeeded or an exception of `ResourceInUse` if
                 profile is referenced by certain clusters/nodes.
        """
        db_profile = self.profile_find(context, identity)
        LOG.info(_LI("Deleting profile '%s'."), identity)
        try:
            profile_base.Profile.delete(context, db_profile.id)
        except exception.EResourceBusy:
            LOG.error(_LI("The profile '%s' cannot be deleted."), identity)
            raise exception.ResourceInUse(resource_type='profile',
                                          resource_id=db_profile.id)

        LOG.info(_LI("Profile '%(id)s' is deleted."), {'id': identity})

    @request_context
    def policy_type_list(self, context):
        """List known policy type implementations.

        :param context: An instance of the request context.
        :return: A list of policy types.
        """
        return environment.global_env().get_policy_types()

    @request_context
    def policy_type_get(self, context, type_name):
        """Get the details about a policy type.

        :param context: An instance of the request context.
        :param type_name: The name of a policy type.
        :return: The details about a policy type.
        """
        policy_type = environment.global_env().get_policy(type_name)
        data = policy_type.get_schema()

        return {
            'name': type_name,
            'schema': data
        }

    def policy_find(self, context, identity, project_safe=True):
        """Find a policy with the given identity.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a profile.
        :param project_safe: A boolean indicating whether policies from
                             projects other than the requesting one should be
                             evaluated.
        :return: A DB object of policy or an exception of `ResourceNotFound`
                 if no matching object is found.
        """
        if uuidutils.is_uuid_like(identity):
            policy = policy_obj.Policy.get(context, identity,
                                           project_safe=project_safe)
            if not policy:
                policy = policy_obj.Policy.get_by_name(
                    context, identity, project_safe=project_safe)
        else:
            policy = policy_obj.Policy.get_by_name(context, identity,
                                                   project_safe=project_safe)
            if not policy:
                policy = policy_obj.Policy.get_by_short_id(
                    context, identity, project_safe=project_safe)

        if not policy:
            raise exception.ResourceNotFound(type='policy', id=identity)

        return policy

    @request_context
    def policy_list(self, context, limit=None, marker=None, sort=None,
                    filters=None, project_safe=True):
        """List policies matching the specified criteria.

        :param context: An instance of request context.
        :param limit: An integer specifying the maximum number of policies to
                      return in a response.
        :param marker: An UUID specifying the policy after which the result
                       list starts.
        :param sort: A list of sorting keys (each optionally attached with a
                     sorting direction) separated by commas.
        :param filters: A dictionary of key-value pairs for filtering out the
                        result list.
        :param project_safe: A boolean indicating whether policies from all
                             projects will be returned.
        :return: A list of `Policy` object representations.
        """
        limit = utils.parse_int_param('limit', limit)
        utils.validate_sort_param(sort, consts.POLICY_SORT_KEYS)
        project_safe = utils.parse_bool_param('project_safe', project_safe)
        if not project_safe and not context.is_admin:
            raise exception.Forbidden()
        policies = policy_base.Policy.load_all(context,
                                               limit=limit, marker=marker,
                                               sort=sort, filters=filters,
                                               project_safe=project_safe)
        return [p.to_dict() for p in policies]

    @request_context
    def _validate_policy(self, context, spec, name=None, validate_props=False):
        """Validate a policy.

        :param context: An instance of the request context.
        :param spec: A dictionary containing the spec for the policy.
        :param name: The name for the policy to be created.
        :param validate_props: Whether to validate the value of property.
        :return: Validated policy object.
        """

        type_name, version = schema.get_spec_version(spec)
        type_str = "-".join([type_name, version])
        try:
            plugin = environment.global_env().get_policy(type_str)
        except exception.ResourceNotFound as ex:
            msg = ex.enhance_msg('specified', ex)
            raise exception.SpecValidationFailed(message=msg)

        kwargs = {
            'user': context.user,
            'project': context.project,
            'domain': context.domain,
        }
        if name is None:
            name = 'validated_policy'
        policy = plugin(name, spec, **kwargs)

        try:
            policy.validate(context, validate_props=validate_props)
        except exception.InvalidSpec as ex:
            msg = six.text_type(ex)
            LOG.error(_LE("Failed in validating policy: %s"), msg)
            raise exception.SpecValidationFailed(message=msg)

        return policy

    @request_context
    def policy_create(self, context, name, spec):
        """Create a policy with the given name and spec.

        :param context: An instance of the request context.
        :param name: The name for the policy to be created.
        :param spec: A dictionary containing the spec for the policy.
        :return: A dictionary containing the details of the policy object
                 created.
        """
        if cfg.CONF.name_unique:
            if policy_obj.Policy.get_by_name(context, name):
                msg = _("A policy named '%(name)s' already exists."
                        ) % {"name": name}
                raise exception.BadRequest(msg=msg)

        policy = self._validate_policy(context, spec, name=name)

        LOG.info(_LI("Creating policy %(type)s '%(name)s'"),
                 {'type': policy.type, 'name': policy.name})

        policy.store(context)
        LOG.info(_LI("Policy '%(name)s' is created: %(id)s."),
                 {'name': name, 'id': policy.id})
        return policy.to_dict()

    @request_context
    def policy_get(self, context, identity):
        """Retrieve the details about a policy.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a policy.
        :return: A dictionary containing the policy details, or an exception
                 of type `ResourceNotFound` if no matching object is found.
        """
        db_policy = self.policy_find(context, identity)
        policy = policy_base.Policy.load(context, db_policy=db_policy)
        return policy.to_dict()

    @request_context
    def policy_update(self, context, identity, name):
        """Update the properties of a given policy.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a policy.
        :param name: The new name for the policy.
        :returns: A dictionary containing the details of the updated policy or
                  an exception `ResourceNotFound` if no matching poicy is
                  found, or an exception `BadRequest` if name is not provided.
        """
        if not name:
            msg = _('Policy name not specified.')
            raise exception.BadRequest(msg=msg)

        db_policy = self.policy_find(context, identity)
        policy = policy_base.Policy.load(context, db_policy=db_policy)

        if name != policy.name:
            LOG.info(_LI("Updating policy '%s'."), identity)
            policy.name = name
            policy.store(context)
            LOG.info(_LI("Policy '%s' is updated."), identity)

        return policy.to_dict()

    @request_context
    def policy_delete(self, context, identity):
        """Delete the specified policy.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a policy.
        :return: None if succeeded or an exception of `ResourceInUse` if
                 policy is still attached to certain clusters.
        """
        db_policy = self.policy_find(context, identity)
        LOG.info(_LI("Delete policy '%s'."), identity)
        try:
            policy_base.Policy.delete(context, db_policy.id)
        except exception.EResourceBusy:
            LOG.error(_LI("Policy '%s' cannot be deleted."), identity)
            raise exception.ResourceInUse(resource_type='policy',
                                          resource_id=db_policy.id)

        LOG.info(_LI("Policy '%s' is deleted."), identity)

    @request_context
    def policy_validate(self, context, spec):
        """Validate a policy with the given properties.

        :param context: An instance of the request context.
        :param spec: A dictionary containing the spec for the policy.
        :return: A dictionary containing the details of the policy object
                 validated.
        """
        policy = self._validate_policy(context, spec, validate_props=True)

        return policy.to_dict()

    def cluster_find(self, context, identity, project_safe=True):
        """Find a cluster with the given identity.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short ID of a cluster.
        :param project_safe: A boolean parameter specifying whether only
                             clusters from the same project are qualified to
                             be returned.
        :return: An instance of `Cluster` class.
        :raises: `ResourceNotFound` if no matching object can be found.
        """

        if uuidutils.is_uuid_like(identity):
            cluster = cluster_obj.Cluster.get(context, identity,
                                              project_safe=project_safe)
            if not cluster:
                cluster = cluster_obj.Cluster.get_by_name(
                    context, identity, project_safe=project_safe)
        else:
            cluster = cluster_obj.Cluster.get_by_name(
                context, identity, project_safe=project_safe)
            # maybe it is a short form of UUID
            if not cluster:
                cluster = cluster_obj.Cluster.get_by_short_id(
                    context, identity, project_safe=project_safe)

        if not cluster:
            raise exception.ResourceNotFound(type='cluster', id=identity)

        return cluster

    @request_context
    def cluster_list(self, context, limit=None, marker=None, sort=None,
                     filters=None, project_safe=True):
        """List clusters matching the specified criteria.

        :param context: An instance of request context.
        :param limit: An integer specifying the maximum number of objects to
                      return in a response.
        :param marker: An UUID specifying the cluster after which the result
                       list starts.
        :param sort: A list of sorting keys (each optionally attached with a
                     sorting direction) separated by commas.
        :param filters: A dictionary of key-value pairs for filtering out the
                        result list.
        :param project_safe: A boolean indicating whether clusters from all
                             projects will be returned.
        :return: A list of `Cluster` object representations.
        """
        limit = utils.parse_int_param('limit', limit)
        utils.validate_sort_param(sort, consts.CLUSTER_SORT_KEYS)
        project_safe = utils.parse_bool_param('project_safe', project_safe)
        if not project_safe and not context.is_admin:
            raise exception.Forbidden()
        clusters = cluster_mod.Cluster.load_all(context, limit=limit,
                                                marker=marker, sort=sort,
                                                filters=filters,
                                                project_safe=project_safe)

        return [cluster.to_dict() for cluster in clusters]

    @request_context
    def cluster_get(self, context, identity):
        """Retrieve the cluster specified.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-ID of a cluster.
        :return: A dictionary containing the details about a cluster.
        """
        db_cluster = self.cluster_find(context, identity)
        cluster = cluster_mod.Cluster.load(context, dbcluster=db_cluster)
        return cluster.to_dict()

    def check_cluster_quota(self, context):
        """Validate the number of clusters created in a project.

        :param context: An instance of the request context.
        :return: None if cluster creation is okay, or an exception of type
                 `Forbbiden` if number of clusters reaches the maximum.
        """
        existing = cluster_obj.Cluster.count_all(context)
        maximum = cfg.CONF.max_clusters_per_project
        if existing >= maximum:
            raise exception.Forbidden()

    @request_context
    def cluster_create(self, context, name, desired_capacity, profile_id,
                       min_size=None, max_size=None, metadata=None,
                       timeout=None):
        """Create a cluster.

        :param context: An instance of the request context.
        :param name: A string specifying the name of the cluster to be created.
        :param desired_capacity: The desired capacity of the cluster.
        :param profile_ID: The UUID, name or short-ID of the profile to use.
        :param min_size: An integer specifying the minimum size of the cluster.
        :param max_size: An integer specifying the maximum size of the cluster.
        :param metadata: A dictionary containing key-value pairs to be
                         associated with the cluster.
        :param timeout: An optional integer specifying the operation timeout
                        value in seconds.
        :return: A dictionary containing the details about the cluster and the
                 ID of the action triggered by this operation.
        """
        self.check_cluster_quota(context)

        if cfg.CONF.name_unique:
            if cluster_obj.Cluster.get_by_name(context, name):
                msg = _("The cluster (%(name)s) already exists."
                        ) % {"name": name}
                raise exception.BadRequest(msg=msg)

        try:
            db_profile = self.profile_find(context, profile_id)
        except exception.ResourceNotFound as ex:
            msg = ex.enhance_msg('specified', ex)
            raise exception.BadRequest(msg=msg)

        init_size = utils.parse_int_param(consts.CLUSTER_DESIRED_CAPACITY,
                                          desired_capacity)
        if min_size is not None:
            min_size = utils.parse_int_param(consts.CLUSTER_MIN_SIZE, min_size)
        if max_size is not None:
            max_size = utils.parse_int_param(consts.CLUSTER_MAX_SIZE, max_size,
                                             allow_negative=True)
        if timeout is not None:
            timeout = utils.parse_int_param(consts.CLUSTER_TIMEOUT, timeout)
        else:
            timeout = cfg.CONF.default_action_timeout

        res = su.check_size_params(None, init_size, min_size, max_size, True)
        if res:
            raise exception.BadRequest(msg=res)

        LOG.info(_LI("Creating cluster '%s'."), name)

        kwargs = {
            'min_size': min_size,
            'max_size': max_size,
            'timeout': timeout,
            'metadata': metadata,
            'user': context.user,
            'project': context.project,
            'domain': context.domain,
        }

        cluster = cluster_mod.Cluster(name, init_size, db_profile.id, **kwargs)
        cluster.store(context)

        # Build an Action for cluster creation
        kwargs = {
            'name': 'cluster_create_%s' % cluster.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
        }
        action_id = action_mod.Action.create(context, cluster.id,
                                             consts.CLUSTER_CREATE, **kwargs)
        dispatcher.start_action()
        LOG.info(_LI("Cluster create action queued: %s."), action_id)

        result = cluster.to_dict()
        result['action'] = action_id
        return result

    @request_context
    def cluster_update(self, context, identity, name=None, profile_id=None,
                       metadata=None, timeout=None):
        """Update a cluster.

        :param context: An instance of the request context.
        :param identity: The UUID, name, or short-ID or the target cluster.
        :param name: A string specifying the new name of the cluster.
        :param profile_id: The UUID, name or short-ID of the new profile.
        :param metadata: A dictionary containing key-value pairs to be
                         associated with the cluster.
        :param timeout: An optional integer specifying the new operation
                        timeout value in seconds.
        :return: A dictionary containing the details about the cluster and the
                 ID of the action triggered by this operation.
        """

        # Get the database representation of the existing cluster
        db_cluster = self.cluster_find(context, identity)
        cluster = cluster_mod.Cluster.load(context, dbcluster=db_cluster)
        if cluster.status == cluster.ERROR:
            msg = _('Updating a cluster in error state')
            LOG.error(msg)
            raise exception.FeatureNotSupported(feature=msg)

        LOG.info(_LI("Updating cluster '%s'."), identity)

        inputs = {}
        if profile_id is not None:
            old_profile = self.profile_find(context, cluster.profile_id)
            try:
                new_profile = self.profile_find(context, profile_id)
            except exception.ResourceNotFound as ex:
                msg = ex.enhance_msg('specified', ex)
                raise exception.BadRequest(msg=msg)

            if new_profile.type != old_profile.type:
                msg = _('Cannot update a cluster to a different profile type, '
                        'operation aborted.')
                raise exception.ProfileTypeNotMatch(message=msg)
            if old_profile.id != new_profile.id:
                inputs['new_profile_id'] = new_profile.id

        if metadata is not None and metadata != cluster.metadata:
            inputs['metadata'] = metadata

        if timeout is not None:
            timeout = utils.parse_int_param(consts.CLUSTER_TIMEOUT, timeout)
            inputs['timeout'] = timeout

        if name is not None:
            inputs['name'] = name

        if not inputs:
            msg = _("No property needs an update.")
            raise exception.BadRequest(msg=msg)

        kwargs = {
            'name': 'cluster_update_%s' % cluster.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': inputs,
        }
        action_id = action_mod.Action.create(context, cluster.id,
                                             consts.CLUSTER_UPDATE, **kwargs)
        dispatcher.start_action()
        LOG.info(_LI("Cluster update action queued: %s."), action_id)

        resp = cluster.to_dict()
        resp['action'] = action_id
        return resp

    @request_context
    def cluster_delete(self, context, identity):
        """Delete the specified cluster.

        :param identity: The UUID, name or short-ID of the target cluster.
        :return: A dictionary containing the ID of the action triggered.
        """

        LOG.info(_LI('Deleting cluster %s'), identity)

        # 'cluster' below is a DB object.
        cluster = self.cluster_find(context, identity)

        if cluster.status in [cluster_mod.Cluster.CREATING,
                              cluster_mod.Cluster.UPDATING,
                              cluster_mod.Cluster.DELETING,
                              cluster_mod.Cluster.RECOVERING]:
            raise exception.ActionInProgress(type='cluster', id=identity,
                                             status=cluster.status)

        containers = cluster.dependents.get('containers', None)
        if containers is not None and len(containers) > 0:
            raise exception.ResourceInUse(resource_type='host_cluster',
                                          resource_id=cluster.id)

        policies = cp_obj.ClusterPolicy.get_all(context, cluster.id)
        if len(policies) > 0:
            msg = _('Cluster %(id)s cannot be deleted without having all '
                    'policies detached.') % {'id': identity}
            LOG.error(msg)
            reason = _("there is still policy(s) attached to it.")
            raise exception.ClusterBusy(cluster=identity, reason=reason)

        receivers = receiver_obj.Receiver.get_all(
            context, filters={'cluster_id': cluster.id})
        if len(receivers) > 0:
            msg = _('Cluster %(id)s cannot be deleted without having all '
                    'receivers deleted.') % {'id': identity}
            LOG.error(msg)
            reason = _("there is still receiver(s) associated with it.")
            raise exception.ClusterBusy(cluster=identity, reason=reason)

        params = {
            'name': 'cluster_delete_%s' % cluster.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
        }
        action_id = action_mod.Action.create(context, cluster.id,
                                             consts.CLUSTER_DELETE, **params)
        dispatcher.start_action()
        LOG.info(_LI("Cluster delete action queued: %s"), action_id)

        return {'action': action_id}

    @request_context
    def cluster_add_nodes(self, context, identity, nodes):
        """Add specified nodes to the specified cluster.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of the target cluster.
        :param nodes: A list of node identities where each item is the UUID,
                      name or short-id of a node.
        :return: A dictionary containing the ID of the action triggered.
        """
        LOG.info(_LI("Adding nodes '%(nodes)s' to cluster '%(cluster)s'."),
                 {'cluster': identity, 'nodes': nodes})

        db_cluster = self.cluster_find(context, identity)
        db_cluster_profile = self.profile_find(context,
                                               db_cluster.profile_id)
        cluster_profile_type = db_cluster_profile.type

        found = []
        not_found = []
        bad_nodes = []
        owned_nodes = []
        not_match_nodes = []
        for node in nodes:
            try:
                db_node = self.node_find(context, node)
                # Skip node in the same cluster already
                if db_node.status != node_mod.Node.ACTIVE:
                    bad_nodes.append(db_node.id)
                elif len(db_node.cluster_id) != 0:
                    owned_nodes.append(db_node.id)
                else:
                    # check profile type matching
                    db_node_profile = self.profile_find(context,
                                                        db_node.profile_id)
                    node_profile_type = db_node_profile.type
                    if node_profile_type != cluster_profile_type:
                        not_match_nodes.append(db_node.id)
                    else:
                        found.append(db_node.id)
            except exception.ResourceNotFound:
                not_found.append(node)
                pass

        error = None
        if len(not_match_nodes) > 0:
            error = _("Profile type of nodes %s does not match that of the "
                      "cluster.") % not_match_nodes
            LOG.error(error)
            raise exception.ProfileTypeNotMatch(message=error)
        elif len(owned_nodes) > 0:
            error = _("Nodes %s already owned by some cluster.") % owned_nodes
            LOG.error(error)
            raise exception.NodeNotOrphan(message=error)
        elif len(bad_nodes) > 0:
            error = _("Nodes are not ACTIVE: %s.") % bad_nodes
        elif len(not_found) > 0:
            error = _("Nodes not found: %s.") % not_found
        elif len(found) == 0:
            error = _("No nodes to add: %s.") % nodes

        if error is not None:
            LOG.error(error)
            raise exception.BadRequest(msg=error)

        target_size = db_cluster.desired_capacity + len(found)
        error = su.check_size_params(db_cluster, target_size, strict=True)
        if error:
            LOG.error(error)
            raise exception.BadRequest(msg=error)

        params = {
            'name': 'cluster_add_nodes_%s' % db_cluster.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': {'nodes': found},
        }
        action_id = action_mod.Action.create(context, db_cluster.id,
                                             consts.CLUSTER_ADD_NODES,
                                             **params)
        dispatcher.start_action()
        LOG.info(_LI("Cluster add nodes action queued: %s."), action_id)

        return {'action': action_id}

    @request_context
    def cluster_del_nodes(self, context, identity, nodes):
        """Delete specified nodes from the named cluster.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of the cluster.
        :param nodes: A list containing the identities of the nodes to delete.
        :return: A dictionary containing the ID of the action triggered.
        """
        LOG.info(_LI("Deleting nodes '%(nodes)s' from cluster '%(cluster)s'."),
                 {'cluster': identity, 'nodes': nodes})
        db_cluster = self.cluster_find(context, identity)
        found = []
        not_found = []
        bad_nodes = []
        for node in nodes:
            try:
                db_node = self.node_find(context, node)
                if db_node.cluster_id != db_cluster.id:
                    bad_nodes.append(db_node.id)
                else:
                    found.append(db_node.id)
            except exception.ResourceNotFound:
                not_found.append(node)
                pass

        error = None
        if len(not_found):
            error = _("Nodes not found: %s.") % not_found
        elif len(bad_nodes):
            error = _("Nodes not members of specified cluster: "
                      "%s.") % bad_nodes
        elif len(found) == 0:
            error = _("No nodes specified.")

        if error is not None:
            LOG.error(error)
            raise exception.BadRequest(msg=error)

        target_size = db_cluster.desired_capacity - len(found)
        error = su.check_size_params(db_cluster, target_size, strict=True)
        if error:
            LOG.error(error)
            raise exception.BadRequest(msg=error)

        params = {
            'name': 'cluster_del_nodes_%s' % db_cluster.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': {
                'candidates': found,
                'count': len(found),
            },
        }
        action_id = action_mod.Action.create(context, db_cluster.id,
                                             consts.CLUSTER_DEL_NODES,
                                             **params)
        dispatcher.start_action()
        LOG.info(_LI("Cluster delete nodes action queued: %s."), action_id)

        return {'action': action_id}

    @request_context
    def cluster_resize(self, context, identity, adj_type=None, number=None,
                       min_size=None, max_size=None, min_step=None,
                       strict=True):
        """Adjust cluster size parameters.

        :param identity: cluster identity which can be cluster name, UUID or
                         short ID.
        :param adj_type: type of adjustment. If specified, must be one of the
                         strings defined in `consts.ADJUSTMENT_TYPES`.
        :param number: number for adjustment. It is interpreted as the new
                       desired_capacity of the cluster if `adj_type` is set
                       to `EXACT_CAPACITY`; it is interpreted as the relative
                       number of nodes to add/remove when `adj_type` is set
                       to `CHANGE_IN_CAPACITY`; it is treated as a percentage
                       when `adj_type` is set to `CHANGE_IN_PERCENTAGE`.
        :param min_size: new lower bound of the cluster size, if specified.
        :param max_size: new upper bound of the cluster size, if specified.
                         A value of negative means no upper limit is imposed.
        :param min_step: the number of nodes to be added or removed when
                         `adj_type` is set to value `CHANGE_IN_PERCENTAGE`
                         and the number calculated is less than 1.
        :param strict: whether Senlin should try a best-effort style resizing
                       or just rejects the request when scaling beyond its
                       current size constraint.

        :return: A dict containing the ID of an action fired.
        """

        # check adj_type
        if adj_type is not None:
            if adj_type not in consts.ADJUSTMENT_TYPES:
                raise exception.InvalidParameter(
                    name=consts.ADJUSTMENT_TYPE, value=adj_type)
            if number is None:
                msg = _('Missing number value for size adjustment.')
                raise exception.BadRequest(msg=msg)
        else:
            if number is not None:
                msg = _('Missing adjustment_type value for size adjustment.')
                raise exception.BadRequest(msg=msg)

        if adj_type == consts.EXACT_CAPACITY:
            number = utils.parse_int_param(consts.ADJUSTMENT_NUMBER, number)
        elif adj_type == consts.CHANGE_IN_CAPACITY:
            number = utils.parse_int_param(consts.ADJUSTMENT_NUMBER, number,
                                           allow_negative=True)
        elif adj_type == consts.CHANGE_IN_PERCENTAGE:
            try:
                number = float(number)
            except ValueError:
                raise exception.InvalidParameter(name=consts.ADJUSTMENT_NUMBER,
                                                 value=number)
            # min_step is only used (so checked) for this case
            if min_step is not None:
                min_step = utils.parse_int_param(consts.ADJUSTMENT_MIN_STEP,
                                                 min_step)

        if min_size is not None:
            min_size = utils.parse_int_param(consts.ADJUSTMENT_MIN_SIZE,
                                             min_size)
        if max_size is not None:
            max_size = utils.parse_int_param(consts.ADJUSTMENT_MAX_SIZE,
                                             max_size, allow_negative=True)
        if strict is not None:
            strict = utils.parse_bool_param(consts.ADJUSTMENT_STRICT, strict)

        db_cluster = self.cluster_find(context, identity)
        current = db_cluster.desired_capacity
        if adj_type is not None:
            desired = su.calculate_desired(current, adj_type, number, min_step)
        else:
            desired = None

        res = su.check_size_params(db_cluster, desired, min_size, max_size,
                                   strict)
        if res:
            raise exception.BadRequest(msg=res)

        fmt = _LI("Resizing cluster '%(cluster)s': type=%(adj_type)s, "
                  "number=%(number)s, min_size=%(min_size)s, "
                  "max_size=%(max_size)s, min_step=%(min_step)s, "
                  "strict=%(strict)s.")
        LOG.info(fmt, {'cluster': identity, 'adj_type': adj_type,
                       'number': number, 'min_size': min_size,
                       'max_size': max_size, 'min_step': min_step,
                       'strict': strict})

        params = {
            'name': 'cluster_resize_%s' % db_cluster.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': {
                consts.ADJUSTMENT_TYPE: adj_type,
                consts.ADJUSTMENT_NUMBER: number,
                consts.ADJUSTMENT_MIN_SIZE: min_size,
                consts.ADJUSTMENT_MAX_SIZE: max_size,
                consts.ADJUSTMENT_MIN_STEP: min_step,
                consts.ADJUSTMENT_STRICT: strict
            }
        }
        action_id = action_mod.Action.create(context, db_cluster.id,
                                             consts.CLUSTER_RESIZE, **params)
        dispatcher.start_action()
        LOG.info(_LI("Cluster resize action queued: %s."), action_id)

        return {'action': action_id}

    @request_context
    def cluster_scale_out(self, context, identity, count=None):
        """Inflate the size of a cluster by then given number (optional).

        :param context: Request context for the call.
        :param identity: The name, ID or short ID of a cluster.
        :param count: The number of nodes to add to the cluster. When omitted,
            a policy gets a chance to decide the count number. When specified,
            a policy would have to respect this input.

        :return: A dict with the ID of the action fired.
        """

        # Validation
        db_cluster = self.cluster_find(context, identity)
        if count is not None:
            count = utils.parse_int_param('count', count, allow_zero=False)
            err = su.check_size_params(db_cluster,
                                       db_cluster.desired_capacity + count)
            if err:
                raise exception.BadRequest(msg=err)

            LOG.info(_LI('Scaling out cluster %(name)s by %(delta)s nodes'),
                     {'name': identity, 'delta': count})
            inputs = {'count': count}
        else:
            LOG.info(_LI('Scaling out cluster %s'), db_cluster.name)
            inputs = {}

        params = {
            'name': 'cluster_scale_out_%s' % db_cluster.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': inputs,
        }
        action_id = action_mod.Action.create(context, db_cluster.id,
                                             consts.CLUSTER_SCALE_OUT,
                                             **params)
        dispatcher.start_action()
        LOG.info(_LI("Cluster Scale out action queued: %s"), action_id)

        return {'action': action_id}

    @request_context
    def cluster_scale_in(self, context, identity, count=None):
        """Deflate the size of a cluster by given number (optional).

        :param context: Request context for the call.
        :param identity: The name, ID or short ID of a cluster.
        :param count: The number of nodes to remove from the cluster. When
            omitted, a policy gets a chance to decide the count number. When
            specified, a policy would have to respect this input.

        :return: A dict with the ID of the action fired.
        """

        db_cluster = self.cluster_find(context, identity)

        if count is not None:
            count = utils.parse_int_param('count', count, allow_zero=False)
            err = su.check_size_params(db_cluster,
                                       db_cluster.desired_capacity - count)
            if err:
                raise exception.BadRequest(msg=err)

            LOG.info(_LI('Scaling in cluster %(name)s by %(delta)s nodes'),
                     {'name': identity, 'delta': count})
            inputs = {'count': count}
        else:
            LOG.info(_LI('Scaling in cluster %s'), db_cluster.name)
            inputs = {}

        params = {
            'name': 'cluster_scale_in_%s' % db_cluster.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': inputs,
        }
        action_id = action_mod.Action.create(context, db_cluster.id,
                                             consts.CLUSTER_SCALE_IN,
                                             **params)
        dispatcher.start_action()
        LOG.info(_LI("Cluster Scale in action queued: %s."), action_id)

        return {'action': action_id}

    @request_context
    def cluster_collect(self, context, identity, path, project_safe=True):
        """Collect a certain attribute across a cluster.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a cluster.
        :param path: A JSONPath-alike string containing path for a particular
                     attribute to aggregate.
        :return: A list containing values of attribute collected from all
                 nodes.
        """
        # validate 'path' string and return a parser,
        # The function may raise a BadRequest exception.
        parser = utils.get_path_parser(path)
        cluster = self.cluster_find(context, identity)
        nodes = node_mod.Node.load_all(context, cluster_id=cluster.id,
                                       project_safe=project_safe)
        attrs = []
        for node in nodes:
            info = node.to_dict()
            if node.physical_id:
                info['details'] = node.get_details(context)
            matches = [m.value for m in parser.find(info)]
            if matches:
                attrs.append({'id': node.id, 'value': matches[0]})

        return {'cluster_attributes': attrs}

    @request_context
    def cluster_check(self, context, identity, params=None):
        """Check the status of a cluster.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a cluster.
        :param params: A dictionary containing additional parameters for
                       the check operation.
        :return: A dictionary containing the ID of the action triggered.
        """
        LOG.info(_LI("Checking Cluster '%(cluster)s'."),
                 {'cluster': identity})
        db_cluster = self.cluster_find(context, identity)

        params = {
            'name': 'cluster_check_%s' % db_cluster.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': params,
        }
        action_id = action_mod.Action.create(context, db_cluster.id,
                                             consts.CLUSTER_CHECK, **params)
        dispatcher.start_action()
        LOG.info(_LI("Cluster check action queued: %s."), action_id)

        return {'action': action_id}

    @request_context
    def cluster_recover(self, context, identity, params=None):
        """Recover a cluster to a healthy status.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a cluster.
        :param params: A dictionary containing additional parameters for
                       the check operation.
        :return: A dictionary containing the ID of the action triggered.
        """
        LOG.info(_LI("Recovering cluster '%s'."), identity)
        db_cluster = self.cluster_find(context, identity)

        params = {
            'name': 'cluster_recover_%s' % db_cluster.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': params,
        }
        action_id = action_mod.Action.create(context, db_cluster.id,
                                             consts.CLUSTER_RECOVER, **params)
        dispatcher.start_action()
        LOG.info(_LI("Cluster recover action queued: %s."), action_id)

        return {'action': action_id}

    def node_find(self, context, identity, project_safe=True):
        """Find a node with the given identity.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a node.
        :param project_safe: A boolean indicating whether only nodes from the
                             same project as the requesting one are qualified
                             to be returned.
        :return: A DB object of Node or an exception of `ResourceNotFound` if
                 no matching object is found.
        """
        if uuidutils.is_uuid_like(identity):
            node = node_obj.Node.get(context, identity,
                                     project_safe=project_safe)
            if not node:
                node = node_obj.Node.get_by_name(context, identity,
                                                 project_safe=project_safe)
        else:
            node = node_obj.Node.get_by_name(context, identity,
                                             project_safe=project_safe)
            if not node:
                node = node_obj.Node.get_by_short_id(
                    context, identity, project_safe=project_safe)

        if node is None:
            raise exception.ResourceNotFound(type='node', id=identity)

        return node

    @request_context
    def node_list(self, context, cluster_id=None, filters=None, sort=None,
                  limit=None, marker=None, project_safe=True):
        """List node records matching the specified criteria.

        :param context: An instance of the request context.
        :param cluster_id: An optional parameter specifying the ID of the
                           cluster from which nodes are chosen.
        :param filters: A dictionary of key-value pairs for filtering out the
                        result list.
        :param sort: A list of sorting keys (each optionally attached with a
                     sorting direction) separated by commas.
        :param limit: An integer specifying the maximum number of objects to
                      return in a response.
        :param marker: An UUID specifying the node after which the result
                       list starts.
        :param project_safe: A boolean indicating whether nodes from all
                             projects will be returned.
        :return: A list of `Node` object representations.
        """
        limit = utils.parse_int_param('limit', limit)
        utils.validate_sort_param(sort, consts.NODE_SORT_KEYS)
        project_safe = utils.parse_bool_param('project_safe', project_safe)
        if not project_safe and not context.is_admin:
            raise exception.Forbidden()

        # Maybe the cluster_id is a name or a short ID
        if cluster_id:
            db_cluster = self.cluster_find(context, cluster_id)
            cluster_id = db_cluster.id
        nodes = node_mod.Node.load_all(context, cluster_id=cluster_id,
                                       limit=limit, marker=marker, sort=sort,
                                       filters=filters,
                                       project_safe=project_safe)

        return [node.to_dict() for node in nodes]

    @request_context
    def node_create(self, context, name, profile_id, cluster_id=None,
                    role=None, metadata=None):
        """Create a node with provided properties.

        :param context: An instance of the request context.
        :param name: Name for the node to be created.
        :param profile_id: The ID, name or short-id of the profile to be used.
        :param cluster_id: The ID, name or short-id of the cluster in which
                           the new node will be a member. This could be None
                           if the node is to be a orphan node.
        :param role: The role for the node to play in the cluster.
        :param metadata: A dictionary containing the key-value pairs to be
                         associated with the node.
        :return: A dictionary containing the details about the node to be
                 created along with the ID of the action triggered by this
                 request.
        """
        if cfg.CONF.name_unique:
            if node_obj.Node.get_by_name(context, name):
                msg = _("The node named (%(name)s) already exists."
                        ) % {"name": name}
                raise exception.BadRequest(msg=msg)

        LOG.info(_LI("Creating node '%s'."), name)

        if cluster_id is None:
            cluster_id = ''

        try:
            node_profile = self.profile_find(context, profile_id)
        except exception.ResourceNotFound as ex:
            msg = ex.enhance_msg('specified', ex)
            raise exception.BadRequest(msg=msg)

        index = -1
        if cluster_id:
            try:
                db_cluster = self.cluster_find(context, cluster_id)
            except exception.ResourceNotFound as ex:
                msg = ex.enhance_msg('specified', ex)

                raise exception.BadRequest(msg=msg)

            cluster_id = db_cluster.id
            if node_profile.id != db_cluster.profile_id:
                cluster_profile = self.profile_find(context,
                                                    db_cluster.profile_id)
                if node_profile.type != cluster_profile.type:
                    msg = _('Node and cluster have different profile type, '
                            'operation aborted.')
                    LOG.error(msg)
                    raise exception.ProfileTypeNotMatch(message=msg)
            index = cluster_obj.Cluster.get_next_index(context, cluster_id)

        # Create a node instance
        kwargs = {
            'index': index,
            'role': role,
            'metadata': metadata or {},
            'user': context.user,
            'project': context.project,
            'domain': context.domain,
        }

        node = node_mod.Node(name, node_profile.id, cluster_id, context,
                             **kwargs)
        node.store(context)

        params = {
            'name': 'node_create_%s' % node.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
        }
        action_id = action_mod.Action.create(context, node.id,
                                             consts.NODE_CREATE, **params)
        dispatcher.start_action()
        LOG.info(_LI("Node create action queued: %s."), action_id)

        result = node.to_dict()
        result['action'] = action_id
        return result

    @request_context
    def node_get(self, context, identity, show_details=False):
        """Get the details about a node.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a node.
        :param show_details: Optional parameter indicating whether the details
                             about the physical object should be returned.
        :return: A dictionary containing the detailed information about a node
                 or an exception of `ResourceNotFound` if no matching node
                 could be found.
        """
        db_node = self.node_find(context, identity)
        node = node_mod.Node.load(context, db_node=db_node)
        res = node.to_dict()
        if show_details and node.physical_id:
            res['details'] = node.get_details(context)
        return res

    @request_context
    def node_update(self, context, identity, name=None, profile_id=None,
                    role=None, metadata=None):
        """Update a node with new propertye values.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of the node.
        :param name: Optional string specifying the new name for the node.
        :param profile_id: The UUID, name or short-id of the new profile to
                           be used.
        :param role: The new role for the node, if specified.
        :param metadata: A dictionary of key-value pairs to be associated with
                         the node.
        :return: A dictionary containing the updated representation of the
                 node along with the ID of the action triggered by this
                 request.
        """
        LOG.info(_LI("Updating node '%s'."), identity)

        db_node = self.node_find(context, identity)

        if profile_id:
            try:
                db_profile = self.profile_find(context, profile_id)
            except exception.ResourceNotFound as ex:
                msg = ex.enhance_msg('specified', ex)
                raise exception.BadRequest(msg=msg)
            profile_id = db_profile.id

            # check if profile_type matches
            old_profile = self.profile_find(context, db_node.profile_id)
            if old_profile.type != db_profile.type:
                msg = _('Cannot update a node to a different profile type, '
                        'operation aborted.')
                LOG.error(msg)
                raise exception.ProfileTypeNotMatch(message=msg)

            inputs = {'new_profile_id': profile_id}
        else:
            inputs = {}

        if name is not None and name != db_node.name:
            inputs['name'] = name
        if role is not None and role != db_node.role:
            inputs['role'] = role
        if metadata is not None and metadata != db_node.metadata:
            inputs['metadata'] = metadata

        if not inputs:
            msg = _("No property needs an update.")
            raise exception.BadRequest(msg=msg)

        params = {
            'name': 'node_update_%s' % db_node.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': inputs,
        }
        action_id = action_mod.Action.create(context, db_node.id,
                                             consts.NODE_UPDATE, **params)
        dispatcher.start_action()
        LOG.info(_LI("Node update action is queued: %s."), action_id)

        node = node_mod.Node.load(context, db_node=db_node)
        resp = node.to_dict()
        resp['action'] = action_id

        return resp

    @request_context
    def node_delete(self, context, identity):
        """Delete the specified node.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of the node.
        :return: A dictionary containing the ID of the action triggered by
                 this request.
        """
        LOG.info(_LI('Deleting node %s'), identity)

        node = self.node_find(context, identity)

        if node.status in [node_mod.Node.CREATING, node_mod.Node.UPDATING,
                           node_mod.Node.DELETING, node_mod.Node.RECOVERING]:
            raise exception.ActionInProgress(type='node', id=identity,
                                             status=node.status)

        containers = node.dependents.get('containers', None)
        if containers is not None and len(containers) > 0:
            raise exception.ResourceInUse(resource_type='host_node',
                                          resource_id=node.id)

        params = {
            'name': 'node_delete_%s' % node.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
        }
        action_id = action_mod.Action.create(context, node.id,
                                             consts.NODE_DELETE, **params)
        dispatcher.start_action()
        LOG.info(_LI("Node delete action is queued: %s."), action_id)

        return {'action': action_id}

    @request_context
    def node_check(self, context, identity, params=None):
        """Check the health status of specified node.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of the node.
        :param params: An dictionary providing additional input parameters
                       for the checking operation.
        :return: A dictionary containing the ID of the action triggered by
                 this request.
        """
        LOG.info(_LI("Checking node '%s'."), identity)

        db_node = self.node_find(context, identity)

        kwargs = {
            'name': 'node_check_%s' % db_node.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': params,
        }
        action_id = action_mod.Action.create(context, db_node.id,
                                             consts.NODE_CHECK, **kwargs)
        dispatcher.start_action()
        LOG.info(_LI("Node check action is queued: %s."), action_id)

        return {'action': action_id}

    @request_context
    def node_recover(self, context, identity, params=None):
        """Recover the specified node.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a node.
        :param params: A dictionary containing the optional parameters for
                       the requested recover operation.
        :return: A dictionary containing the ID of the action triggered by the
                 recover request.
        """
        LOG.info(_LI("Recovering node '%s'."), identity)

        db_node = self.node_find(context, identity)

        kwargs = {
            'name': 'node_recover_%s' % db_node.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': params
        }
        action_id = action_mod.Action.create(context, db_node.id,
                                             consts.NODE_RECOVER, **kwargs)
        dispatcher.start_action()
        LOG.info(_LI("Node recover action is queued: %s."), action_id)

        return {'action': action_id}

    @request_context
    def cluster_policy_list(self, context, identity, filters=None, sort=None):
        """List cluster-policy bindings given the cluster identity.

        :param context: An instance of the request context.
        :param identity: The ID, name or short ID of the target cluster.
        :param filters: A list of key-value pairs for filtering out the result
                        list.
        :param sort: A list of sorting keys (optionally appended with sorting
                     directions) separated by commas.
        :return: A list containing dictionaries each representing a binding.
        """
        utils.validate_sort_param(sort, consts.CLUSTER_POLICY_SORT_KEYS)
        db_cluster = self.cluster_find(context, identity)
        bindings = cpm.ClusterPolicy.load_all(
            context, db_cluster.id, filters=filters, sort=sort)

        return [binding.to_dict() for binding in bindings]

    @request_context
    def cluster_policy_get(self, context, identity, policy_id):
        """Get the binding record giving the cluster and policy identity.

        :param context: An instance of the request context.
        :param identity: The ID, name or short ID of the target cluster.
        :param policy_id: The ID, name or short ID of the target policy.
        :return: A dictionary containing the binding record, or raises an
                 exception of ``PolicyNotAttached``.
        """
        db_cluster = self.cluster_find(context, identity)
        db_policy = self.policy_find(context, policy_id)

        try:
            binding = cpm.ClusterPolicy.load(
                context, db_cluster.id, db_policy.id)
        except exception.PolicyNotAttached:
            raise exception.PolicyBindingNotFound(policy=policy_id,
                                                  identity=identity)

        return binding.to_dict()

    @request_context
    def cluster_policy_attach(self, context, identity, policy, enabled=True):
        """Attach a policy to the specified cluster.

        This is done via an action because a cluster lock is needed.

        :param context: An instance of request context.
        :param identity: The ID, name or short ID of the target cluster.
        :param policy: The ID, name or short ID of the target policy.
        :param enabled: Optional parameter specifying whether the policy is
                        enabled when attached.
        :return: A dictionary containg the ID of the action fired.
        """
        LOG.info(_LI("Attaching policy (%(policy)s) to cluster "
                     "(%(cluster)s)."),
                 {'policy': policy, 'cluster': identity})

        db_cluster = self.cluster_find(context, identity)
        try:
            db_policy = self.policy_find(context, policy)
        except exception.ResourceNotFound as ex:
            msg = ex.enhance_msg('specified', ex)
            raise exception.BadRequest(msg=msg)

        params = {
            'name': 'attach_policy_%s' % db_cluster.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': {
                'policy_id': db_policy.id,
                'enabled': utils.parse_bool_param('enabled', enabled) or True,
            }
        }
        action_id = action_mod.Action.create(context, db_cluster.id,
                                             consts.CLUSTER_ATTACH_POLICY,
                                             **params)
        dispatcher.start_action()
        LOG.info(_LI("Policy attach action queued: %s."), action_id)

        return {'action': action_id}

    @request_context
    def cluster_policy_detach(self, context, identity, policy):
        """Detach a policy from the specified cluster.

        This is done via an action because cluster lock is needed.

        :param context: An instance of request context.
        :param identity: The ID, name or short ID of the target cluster.
        :param policy: The ID, name or short ID of the target policy.
        :return: A dictionary containg the ID of the action fired.
        """
        LOG.info(_LI("Detaching policy '%(policy)s' from cluster "
                     "'%(cluster)s'."),
                 {'policy': policy, 'cluster': identity})

        db_cluster = self.cluster_find(context, identity)
        try:
            db_policy = self.policy_find(context, policy)
        except exception.ResourceNotFound as ex:
            msg = ex.enhance_msg('specified', ex)
            raise exception.BadRequest(msg=msg)

        binding = cp_obj.ClusterPolicy.get(context, db_cluster.id,
                                           db_policy.id)
        if binding is None:
            msg = _("The policy (%(p)s) is not attached to the specified "
                    "cluster (%(c)s).") % {'p': policy, 'c': identity}
            raise exception.BadRequest(msg=msg)

        params = {
            'name': 'detach_policy_%s' % db_cluster.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': {'policy_id': db_policy.id},
        }
        action_id = action_mod.Action.create(context, db_cluster.id,
                                             consts.CLUSTER_DETACH_POLICY,
                                             **params)
        dispatcher.start_action()
        LOG.info(_LI("Policy dettach action queued: %s."), action_id)

        return {'action': action_id}

    @request_context
    def cluster_policy_update(self, context, identity, policy, enabled=None):
        """Update an existing policy binding on a cluster.

        This is done via an action because cluster lock is needed.
        :param context: An instance of request context.
        :param identity: The ID, name or short ID of the target cluster.
        :param policy: The ID, name or short ID of the target policy.
        :param enabled: Optional parameter specifying whether the policy is
                        enabled after the update.
        :return: A dictionary containg the ID of the action fired.
        """
        LOG.info(_LI("Updating policy '%(policy)s' on cluster '%(cluster)s.'"),
                 {'policy': policy, 'cluster': identity})

        db_cluster = self.cluster_find(context, identity)
        try:
            db_policy = self.policy_find(context, policy)
        except exception.ResourceNotFound as ex:
            msg = ex.enhance_msg('specified', ex)
            raise exception.BadRequest(msg=msg)

        binding = cp_obj.ClusterPolicy.get(context, db_cluster.id,
                                           db_policy.id)
        if binding is None:
            msg = _("The policy (%(p)s) is not attached to the specified "
                    "cluster (%(c)s).") % {'p': policy, 'c': identity}
            raise exception.BadRequest(msg=msg)

        inputs = {'policy_id': db_policy.id}
        if enabled is not None:
            inputs['enabled'] = utils.parse_bool_param('enabled', enabled)

        params = {
            'name': 'update_policy_%s' % db_cluster.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': inputs
        }
        action_id = action_mod.Action.create(context, db_cluster.id,
                                             consts.CLUSTER_UPDATE_POLICY,
                                             **params)
        dispatcher.start_action()
        LOG.info(_LI("Policy update action queued: %s."), action_id)

        return {'action': action_id}

    def action_find(self, context, identity, project_safe=True):
        """Find an action with the given identity.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of an action.
        :return: A DB object of action or an exception `ResourceNotFound` if
                 no matching action is found.
        """
        if uuidutils.is_uuid_like(identity):
            action = action_obj.Action.get(context, identity,
                                           project_safe=project_safe)
            if not action:
                action = action_obj.Action.get_by_name(
                    context, identity, project_safe=project_safe)
        else:
            action = action_obj.Action.get_by_name(
                context, identity, project_safe=project_safe)
            if not action:
                action = action_obj.Action.get_by_short_id(
                    context, identity, project_safe=project_safe)

        if not action:
            raise exception.ResourceNotFound(type='action', id=identity)

        return action

    @request_context
    def action_list(self, context, filters=None, limit=None, marker=None,
                    sort=None, project_safe=True):
        """List action records matching the specified criteria.

        :param context: An instance of the request context.
        :param filters: A dictionary of key-value pairs for filtering out the
                        result list.
        :param limit: An integer specifying the maximum number of objects to
                      return in a response.
        :param marker: An UUID specifying the action after which the result
                       list starts.
        :param sort: A list of sorting keys (each optionally attached with a
                     sorting direction) separated by commas.
        :param project_safe: A boolean indicating whether actions from all
                             projects will be returned.
        :return: A list of `Action` object representations.
        """
        limit = utils.parse_int_param('limit', limit)
        utils.validate_sort_param(sort, consts.ACTION_SORT_KEYS)
        project_safe = utils.parse_bool_param('project_safe', project_safe)
        if not project_safe and not context.is_admin:
            raise exception.Forbidden()
        results = action_mod.Action.load_all(context, filters=filters,
                                             limit=limit, marker=marker,
                                             sort=sort,
                                             project_safe=project_safe)

        return [a.to_dict() for a in results]

    @request_context
    def action_create(self, context, name, cluster, action, inputs=None):
        """Create an action with given details.

        :param context: Request context instance.
        :param name: Name of the action.
        :param cluster: Name, ID or short ID of the targeted cluster.
        :param action: String representation of the action.
        :param inputs: Optional inputs for the action.
        :return: A dict containing the action created.
        """
        LOG.info(_LI("Creating action '%s'."), name)

        target = self.cluster_find(context, cluster)

        # Create an action instance
        params = {
            'name': name,
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': inputs or {},
        }
        action_id = action_mod.Action.create(context, target.id, action,
                                             **params)

        # TODO(Anyone): Uncomment this to notify the dispatcher
        # dispatcher.start_action(action_id=action.id)

        LOG.info(_LI("Action '%(name)s' is created: %(id)s."),
                 {'name': name, 'id': action_id})
        return {'action': action_id}

    @request_context
    def action_get(self, context, identity):
        """Get the details about specified action.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of an action.
        :return: A dictionary containing the details about an action, or an
                 exception `ResourceNotFound` if no matching action is found.
        """
        db_action = self.action_find(context, identity)
        action = action_mod.Action.load(context, db_action=db_action)
        return action.to_dict()

    @request_context
    def action_delete(self, context, identity):
        """Delete the specified action object.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of an action object.
        :return: None if deletion was successful, or an exception of type
                 `ResourceInUse`.
        """
        db_action = self.action_find(context, identity)
        LOG.info(_LI("Deleting action '%s'."), identity)
        try:
            action_mod.Action.delete(context, db_action.id)
        except exception.EResourceBusy:
            raise exception.ResourceInUse(resource_type='action',
                                          resource_id=db_action.id)

        LOG.info(_LI("Action '%s' is deleted."), identity)

    def receiver_find(self, context, identity, project_safe=True):
        """Find a receiver with the given identity.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a receiver.
        :param project_safe: A boolean indicating whether receiver from other
                             projects other than the requesting one can be
                             returned.
        :return: A DB object of receiver or an exception `ResourceNotFound`
                 if no matching receiver is found.
        """
        if uuidutils.is_uuid_like(identity):
            receiver = receiver_obj.Receiver.get(
                context, identity, project_safe=project_safe)
            if not receiver:
                receiver = receiver_obj.Receiver.get_by_name(
                    context, identity, project_safe=project_safe)
        else:
            receiver = receiver_obj.Receiver.get_by_name(
                context, identity, project_safe=project_safe)
            if not receiver:
                receiver = receiver_obj.Receiver.get_by_short_id(
                    context, identity, project_safe=project_safe)

        if not receiver:
            raise exception.ResourceNotFound(type='receiver', id=identity)

        return receiver

    @request_context
    def receiver_list(self, context, limit=None, marker=None, sort=None,
                      filters=None, project_safe=True):
        """List receivers matching the specified criteria.

        :param context: An instance of the request context.
        :param limit: An integer specifying the maximum number of objects to
                      return in a response.
        :param marker: An UUID specifying the receiver after which the result
                       list starts.
        :param sort: A list of sorting keys (each optionally attached with a
                     sorting direction) separated by commas.
        :param filters: A dictionary of key-value pairs for filtering out the
                        result list.
        :param project_safe: A boolean indicating whether receivers from all
                             projects will be returned.
        :return: A list of `Receiver` object representations.
        """
        limit = utils.parse_int_param('limit', limit)
        utils.validate_sort_param(sort, consts.RECEIVER_SORT_KEYS)
        project_safe = utils.parse_bool_param('project_safe', project_safe)
        if not project_safe and not context.is_admin:
            raise exception.Forbidden()
        receivers = receiver_mod.Receiver.load_all(context, limit=limit,
                                                   marker=marker,
                                                   sort=sort, filters=filters,
                                                   project_safe=project_safe)
        return [r.to_dict() for r in receivers]

    @request_context
    def receiver_create(self, context, name, type_name, cluster_id, action,
                        actor=None, params=None):
        """Create a receiver.

        :param context: An instance of the request context.
        :param name: Name of the receiver.
        :param type_name: Name of the receiver type, subject to validation.
        :param cluster_id: UUID, name or short-id of a cluster.
        :param action: Name or ID of an action, currently only builtin action
                       names are supported.
        :param actor: Future extension.
        :param params: A dictionary containing key-value pairs as inputs to
                       the action.
        :return: A dictionary containing the details about the receiver
                 created.
        """
        if cfg.CONF.name_unique:
            if receiver_obj.Receiver.get_by_name(context, name):
                msg = _("A receiver named '%s' already exists.") % name
                raise exception.BadRequest(msg=msg)

        LOG.info(_LI("Creating receiver %(n)s: \n"
                     "  type: %(t)s\n  cluster: %(c)s\n  action: %(a)s."),
                 {'n': name, 't': type_name, 'c': cluster_id, 'a': action})

        rtype = type_name.lower()
        if rtype not in consts.RECEIVER_TYPES:
            msg = _("Receiver type '%s' is not supported.") % rtype
            raise exception.BadRequest(msg=msg)

        # Sanity check for webhook target
        cluster = None
        if rtype == consts.RECEIVER_WEBHOOK:
            # Check whether cluster identified by cluster_id does exist
            try:
                cluster = self.cluster_find(context, cluster_id)
            except exception.ResourceNotFound as ex:
                msg = ex.enhance_msg('referenced', ex)
                raise exception.BadRequest(msg=msg)

            # permission checking
            if not context.is_admin and context.user != cluster.user:
                raise exception.Forbidden()

            # Check action name
            if action not in consts.ACTION_NAMES:
                msg = _("Illegal action '%s' specified.") % action
                raise exception.BadRequest(msg=msg)

            if action.lower().split('_')[0] != 'cluster':
                msg = _("Action '%s' is not applicable to clusters."
                        ) % action
                raise exception.BadRequest(msg=msg)

        if not params:
            params = {}
        kwargs = {
            'name': name,
            'user': context.user,
            'project': context.project,
            'domain': context.domain,
            'params': params
        }

        receiver = receiver_mod.Receiver.create(context, rtype, cluster,
                                                action, **kwargs)
        LOG.info(_LI("Receiver (%(n)s) is created: %(i)s."),
                 {'n': name, 'i': receiver.id})

        return receiver.to_dict()

    @request_context
    def receiver_get(self, context, identity, project_safe=True):
        """Get the details about a receiver.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a receiver.
        :param project_safe: Whether matching object from other projects can
                             be returned.
        :return: A dictionary containing the details about a receiver or
                 an exception `ResourceNotFound` if no matching object found.
        """
        db_receiver = self.receiver_find(context, identity,
                                         project_safe=project_safe)
        receiver = receiver_mod.Receiver.load(context,
                                              receiver_obj=db_receiver,
                                              project_safe=project_safe)
        return receiver.to_dict()

    @request_context
    def receiver_delete(self, context, identity):
        """Delete the specified receiver.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of a receiver.
        :return: None if successfully deleted the receiver or an exception of
                 `ResourceNotFound` if the object could not be found.
        """
        db_receiver = self.receiver_find(context, identity)
        LOG.info(_LI("Deleting receiver %s."), identity)
        receiver_mod.Receiver.delete(context, db_receiver.id)
        LOG.info(_LI("Receiver %s is deleted."), identity)

    @request_context
    def webhook_trigger(self, context, identity, params=None):

        LOG.info(_LI("Triggering webhook (%s)."), identity)
        receiver = self.receiver_find(context, identity)

        try:
            cluster = self.cluster_find(context, receiver.cluster_id)
        except exception.ResourceNotFound as ex:
            msg = ex.enhance_msg('referenced', ex)
            raise exception.BadRequest(msg=msg)

        data = copy.deepcopy(receiver.params)
        if params:
            data.update(params)

        kwargs = {
            'name': 'webhook_%s' % receiver.id[:8],
            'cause': action_mod.CAUSE_RPC,
            'status': action_mod.Action.READY,
            'inputs': data,
        }
        action_id = action_mod.Action.create(context, cluster.id,
                                             receiver.action, **kwargs)
        dispatcher.start_action()
        LOG.info(_LI("Webhook %(w)s' triggered with action queued: %(a)s."),
                 {'w': identity, 'a': action_id})

        return {'action': action_id}

    def event_find(self, context, identity, project_safe=True):
        """Find an event with the given identity.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of the event.
        :param project_safe: A boolean specifying that only events from the
                             same project as the requesting one are qualified
                             to be returned.
        :return: A dictionary containing the details of the event.
        """
        event = None
        if uuidutils.is_uuid_like(identity):
            event = event_obj.Event.get(context, identity,
                                        project_safe=project_safe)
        if not event:
            event = event_obj.Event.get_by_short_id(context, identity,
                                                    project_safe=project_safe)
        if not event:
            raise exception.ResourceNotFound(type='event', id=identity)

        return event

    @request_context
    def event_list(self, context, filters=None, limit=None, marker=None,
                   sort=None, project_safe=True):
        """List event records matching the specified criteria.

        :param context: An instance of the request context.
        :param filters: A dictionary of key-value pairs for filtering out the
                        result list.
        :param limit: An integer specifying the maximum number of objects to
                      return in a response.
        :param marker: An UUID specifying the event after which the result
                       list starts.
        :param sort: A list of sorting keys (each optionally attached with a
                     sorting direction) separated by commas.
        :param project_safe: A boolean indicating whether events from all
                             projects will be returned.
        :return: A list of `Event` object representations.
        """
        limit = utils.parse_int_param('limit', limit)
        utils.validate_sort_param(sort, consts.EVENT_SORT_KEYS)
        project_safe = utils.parse_bool_param('project_safe', project_safe)
        if not project_safe and not context.is_admin:
            raise exception.Forbidden()

        if filters and consts.EVENT_LEVEL in filters:
            value = filters.pop(consts.EVENT_LEVEL)
            value = utils.parse_level_values(value)
            if value is not None:
                filters[consts.EVENT_LEVEL] = value

        all_events = event_obj.Event.get_all(context, filters=filters,
                                             limit=limit, marker=marker,
                                             sort=sort,
                                             project_safe=project_safe)

        results = [event.as_dict() for event in all_events]
        return results

    @request_context
    def event_get(self, context, identity):
        """Get the details about a specified event.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of an event.
        :return: A dictionary containing the details about the event or an
                 exception of `ResourceNotFound` if no matching record could
                 be found.
        """
        db_event = self.event_find(context, identity)
        return db_event.as_dict()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest.lib import decorators
from tempest.lib import exceptions

from senlin.tests.tempest.api import base


class TestNodeActionNegativeBadRequest(base.BaseSenlinAPITest):

    @decorators.idempotent_id('9ab462e2-ea3a-49f8-bd78-5a056ae80a48')
    def test_node_action_no_action_specified(self):
        # No action is specified
        params = {}

        # Verify badrequest exception(400) is raised.
        self.assertRaises(exceptions.BadRequest,
                          self.client.trigger_action,
                          'nodes', 'node_id', params)

    @decorators.idempotent_id('9e696e4f-1ec8-45d7-8461-81d275aae81d')
    def test_node_action_multiple_action_specified(self):
        # Multiple actions are specified
        params = {
            'check': {},
            'recover': {}
        }

        # Verify badrequest exception(400) is raised.
        self.assertRaises(exceptions.BadRequest,
                          self.client.trigger_action,
                          'nodes', 'node_id', params)

    @decorators.idempotent_id('4bd97c71-fbfc-421d-95ff-b3f4a212cc38')
    def test_node_action_unrecognized_action(self):
        # Unrecoginized action is specified
        params = {
            'bogus': {}
        }

        # Verify badrequest exception(400) is raised.
        self.assertRaises(exceptions.BadRequest,
                          self.client.trigger_action,
                          'nodes', 'node_id', params)

    @decorators.idempotent_id('723ea351-1bcb-4d45-bfe7-35c656d29761')
    def test_node_action_check_param_is_not_map(self):
        # Check action parameter is not a map
        params = {
            'check': []
        }

        # Verify badrequest exception(400) is raised.
        self.assertRaises(exceptions.BadRequest,
                          self.client.trigger_action,
                          'nodes', 'node_id', params)

    @decorators.idempotent_id('60790d8a-fd9e-47d8-b9e2-c06aa7701c33')
    def test_node_action_recover_param_is_not_map(self):
        # Recover action parameter is not a map
        params = {
            'recover': []
        }

        # Verify badrequest exception(400) is raised.
        self.assertRaises(exceptions.BadRequest,
                          self.client.trigger_action,
                          'nodes', 'node_id', params)


class TestNodeActionNegativeNotFound(base.BaseSenlinAPITest):

    @decorators.idempotent_id('90c46123-f992-4833-859a-46f6d2ccd8e9')
    def test_node_action_node_not_found(self):
        params = {
            'check': {}
        }

        # Verify notfound exception(404) is raised.
        self.assertRaises(exceptions.NotFound,
                          self.client.trigger_action,
                          'nodes', '90c46123-f992-4833-859a-46f6d2ccd8e9',
                          params)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import mock
from oslo_utils import timeutils
from oslo_utils import uuidutils

from senlin import objects
from senlin.objects import base
from senlin.objects import fields
from senlin.objects import notification
from senlin.tests.unit.common import base as test_base
from senlin.tests.unit.common import utils


@base.SenlinObjectRegistry.register_if(False)
class TestObject(base.SenlinObject):
    VERSION = '1.0'
    fields = {
        'field_1': fields.StringField(),
        'field_2': fields.IntegerField(),
        'not_important_field': fields.IntegerField(),
    }


@base.SenlinObjectRegistry.register_if(False)
class TestPayload(notification.NotificationPayloadBase):
    VERSION = '1.0'

    schema = {
        'field_1': ('source_field', 'field_1'),
        'field_2': ('source_field', 'field_2'),
    }

    fields = {
        'extra_field': fields.StringField(),  # filled by ctor
        'field_1': fields.StringField(),  # filled by the schema
        'field_2': fields.IntegerField(),   # filled by the schema
    }

    def populate_schema(self, source_field):
        super(TestPayload, self).populate_schema(source_field=source_field)


@base.SenlinObjectRegistry.register_if(False)
class TestPayloadEmptySchema(notification.NotificationPayloadBase):
    VERSION = '1.0'

    fields = {
        'extra_field': fields.StringField(),  # filled by ctor
    }


@base.SenlinObjectRegistry.register_if(False)
class TestNotification(notification.NotificationBase):
    VERSION = '1.0'
    fields = {
        'payload': fields.ObjectField('TestPayload')
    }


@base.SenlinObjectRegistry.register_if(False)
class TestNotificationEmptySchema(notification.NotificationBase):
    VERSION = '1.0'
    fields = {
        'payload': fields.ObjectField('TestPayloadEmptySchema')
    }


class TestNotificationBase(test_base.SenlinTestCase):

    fake_service = {
        'created_at': timeutils.utcnow().replace(microsecond=0),
        'updated_at': None,
        'id': 123,
        'host': 'fake-host',
        'binary': 'senlin-fake',
        'topic': 'fake-service-topic',
        'disabled': False,
        'disabled_reason': None,
    }

    expected_payload = {
        'versioned_object.name': 'TestPayload',
        'versioned_object.data': {
            'extra_field': 'test string',
            'field_1': 'test1',
            'field_2': 42
        },
        'versioned_object.version': '1.0',
        'versioned_object.namespace': 'senlin'
    }

    def setUp(self):
        super(TestNotificationBase, self).setUp()
        self.ctx = utils.dummy_context()
        service_id = uuidutils.generate_uuid()
        self.service_obj = objects.Service.create(
            self.ctx, service_id,
            self.fake_service['host'],
            self.fake_service['binary'],
            self.fake_service['topic'])

        self.my_obj = TestObject(field_1='test1', field_2=42,
                                 not_important_field=13)

        self.payload = TestPayload(extra_field='test string')
        self.payload.populate_schema(source_field=self.my_obj)

        self.notification = TestNotification(
            event_type=notification.EventType(
                object='test_object',
                action=fields.NotificationAction.UPDATE,
                phase=fields.NotificationPhase.START),
            publisher=notification.NotificationPublisher.from_service_obj(
                self.service_obj),
            priority=fields.NotificationPriority.INFO,
            payload=self.payload)

    def _verify_notification(self, mock_notifier, mock_context,
                             expected_event_type, expected_payload):
        mock_notifier.prepare.assert_called_once_with(
            publisher_id='senlin-fake:fake-host')
        mock_notify = mock_notifier.prepare.return_value.info
        self.assertTrue(mock_notify.called)
        self.assertEqual(mock_notify.call_args[0][0], mock_context)
        self.assertEqual(mock_notify.call_args[1]['event_type'],
                         expected_event_type)
        actual_payload = mock_notify.call_args[1]['payload']
        self.assertJsonEqual(expected_payload, actual_payload)

    @mock.patch('senlin.common.messaging.NOTIFIER')
    def test_emit_notification(self, mock_notifier):

        mock_context = mock.Mock()
        mock_context.to_dict.return_value = {}
        self.notification.emit(mock_context)

        self._verify_notification(
            mock_notifier,
            mock_context,
            expected_event_type='test_object.update.start',
            expected_payload=self.expected_payload)

    @mock.patch('senlin.common.messaging.NOTIFIER')
    def test_emit_with_host_and_binary_as_publisher(self, mock_notifier):
        event_type = notification.EventType(
            object='test_object',
            action=fields.NotificationAction.UPDATE)
        publisher = notification.NotificationPublisher(host='fake-host',
                                                       binary='senlin-fake')

        noti = TestNotification(event_type=event_type,
                                publisher=publisher,
                                priority=fields.NotificationPriority.INFO,
                                payload=self.payload)

        mock_context = mock.Mock()
        mock_context.to_dict.return_value = {}
        noti.emit(mock_context)

        self._verify_notification(
            mock_notifier,
            mock_context,
            expected_event_type='test_object.update',
            expected_payload=self.expected_payload)

    @mock.patch('senlin.common.messaging.NOTIFIER')
    def test_emit_event_type_without_phase(self, mock_notifier):
        noti = TestNotification(
            event_type=notification.EventType(
                object='test_object',
                action=fields.NotificationAction.UPDATE),
            publisher=notification.NotificationPublisher.from_service_obj(
                self.service_obj),
            priority=fields.NotificationPriority.INFO,
            payload=self.payload)

        mock_context = mock.Mock()
        mock_context.to_dict.return_value = {}
        noti.emit(mock_context)

        self._verify_notification(
            mock_notifier,
            mock_context,
            expected_event_type='test_object.update',
            expected_payload=self.expected_payload)

    @mock.patch('senlin.common.messaging.NOTIFIER')
    def test_not_possible_to_emit_if_not_populated(self, mock_notifier):
        # create a non-populated payload
        payload = TestPayload(extra_field='test string')
        event_type = notification.EventType(
            object='test_object',
            action=fields.NotificationAction.UPDATE)
        publisher = notification.NotificationPublisher.from_service_obj(
            self.service_obj)

        noti = TestNotification(
            event_type=event_type,
            publisher=publisher,
            priority=fields.NotificationPriority.INFO,
            payload=payload)

        mock_context = mock.Mock()
        self.assertRaises(AssertionError, noti.emit, mock_context)
        self.assertFalse(mock_notifier.called)

    @mock.patch('senlin.common.messaging.NOTIFIER')
    def test_empty_schema(self, mock_notifier):
        # create a non-populated payload
        payload = TestPayloadEmptySchema(extra_field='test string')
        event_type = notification.EventType(
            object='test_object',
            action=fields.NotificationAction.UPDATE)
        publisher = notification.NotificationPublisher.from_service_obj(
            self.service_obj)

        noti = TestNotificationEmptySchema(
            event_type=event_type,
            publisher=publisher,
            priority=fields.NotificationPriority.INFO,
            payload=payload)

        mock_context = mock.Mock()
        mock_context.to_dict.return_value = {}
        noti.emit(mock_context)

        self._verify_notification(
            mock_notifier,
            mock_context,
            expected_event_type='test_object.update',
            expected_payload={
                'versioned_object.name': 'TestPayloadEmptySchema',
                'versioned_object.data': {'extra_field': u'test string'},
                'versioned_object.version': '1.0',
                'versioned_object.namespace': 'senlin'
            }
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import mock
import socket

from keystoneauth1 import loading as ks_loading
from oslo_config import cfg

from senlin.common import exception
from senlin.drivers import base as driver_base
from senlin.engine.receivers import message as mmod
from senlin.tests.unit.common import base

UUID = 'aa5f86b8-e52b-4f2b-828a-4c14c770938d'


class TestMessage(base.SenlinTestCase):

    @mock.patch.object(driver_base, 'SenlinDriver')
    def test_keystone_client(self, mock_senlindriver):
        sd = mock.Mock()
        kc = mock.Mock()
        sd.identity.return_value = kc
        mock_senlindriver.return_value = sd

        message = mmod.Message('message', None, None, user='user1',
                               project='project1')

        # cached will be returned
        message._keystoneclient = kc
        self.assertEqual(kc, message.keystone())

        # new keystone client created if no cache found
        message._keystoneclient = None
        params = mock.Mock()
        mock_param = self.patchobject(mmod.Message, '_build_conn_params',
                                      return_value=params)
        res = message.keystone()
        self.assertEqual(kc, res)
        self.assertEqual(kc, message._keystoneclient)
        mock_param.assert_called_once_with('user1', 'project1')
        sd.identity.assert_called_once_with(params)

    @mock.patch.object(driver_base, 'SenlinDriver')
    def test_zaqar_client(self, mock_senlindriver):
        sd = mock.Mock()
        zc = mock.Mock()
        sd.message.return_value = zc
        mock_senlindriver.return_value = sd

        message = mmod.Message('message', None, None, user='user1',
                               project='project1')

        # cached will be returned
        message._zaqarclient = zc
        self.assertEqual(zc, message.zaqar())

        # new zaqar client created if no cache found
        message._zaqarclient = None
        params = mock.Mock()
        mock_param = self.patchobject(mmod.Message, '_build_conn_params',
                                      return_value=params)
        res = message.zaqar()
        self.assertEqual(zc, res)
        self.assertEqual(zc, message._zaqarclient)
        mock_param.assert_called_once_with('user1', 'project1')
        sd.message.assert_called_once_with(params)

    def test__generate_subscriber_url_host_provided(self):
        cfg.CONF.set_override('host', 'web.com', 'webhook')
        cfg.CONF.set_override('port', '1234', 'webhook')
        message = mmod.Message('message', None, None, id=UUID)
        res = message._generate_subscriber_url()

        expected = 'trust+http://web.com:1234/v1/receivers/%s/notify' % UUID
        self.assertEqual(expected, res)

    @mock.patch.object(mmod.Message, '_get_base_url')
    def test__generate_subscriber_url_host_not_provided(
            self, mock_get_base_url):
        mock_get_base_url.return_value = 'http://web.com:1234/v1'
        message = mmod.Message('message', None, None, id=UUID)
        res = message._generate_subscriber_url()

        expected = 'trust+http://web.com:1234/v1/receivers/%s/notify' % UUID
        self.assertEqual(expected, res)

    @mock.patch.object(socket, 'gethostname')
    @mock.patch.object(mmod.Message, '_get_base_url')
    def test__generate_subscriber_url_no_host_no_base(
            self, mock_get_base_url, mock_gethostname):
        mock_get_base_url.return_value = None
        mock_gethostname.return_value = 'test-host'
        message = mmod.Message('message', None, None, id=UUID)
        res = message._generate_subscriber_url()

        expected = 'trust+http://test-host:8778/v1/receivers/%s/notify' % UUID
        self.assertEqual(expected, res)

    def test_to_dict(self):
        message = mmod.Message('message', None, None, user='user1',
                               project='project1', id=UUID)
        message.channel = {'queue_name': 'test-queue',
                           'subscription': 'subscription-id'}
        res = message.to_dict()
        expected_res = {
            'name': None,
            'id': UUID,
            'user': 'user1',
            'project': 'project1',
            'domain': '',
            'type': 'message',
            'channel': {'queue_name': 'test-queue'},
            'action': None,
            'cluster_id': None,
            'actor': {},
            'params': {},
            'created_at': None,
            'updated_at': None,
        }
        self.assertEqual(expected_res, res)

    @mock.patch.object(mmod.Message, '_create_queue')
    @mock.patch.object(mmod.Message, '_create_subscription')
    def test_initialize_channel(self, mock_create_subscription,
                                mock_create_queue):
        mock_sub = mock.Mock()
        mock_sub.subscription_id = 'test-subscription-id'
        mock_create_subscription.return_value = mock_sub
        mock_create_queue.return_value = 'test-queue'

        message = mmod.Message('message', None, None)
        res = message.initialize_channel()

        expected_channel = {'queue_name': 'test-queue',
                            'subscription': 'test-subscription-id'}
        self.assertEqual(expected_channel, res)
        mock_create_queue.assert_called_once_with()
        mock_create_subscription.assert_called_once_with('test-queue')

    @mock.patch.object(mmod.Message, 'zaqar')
    def test__create_queue(self, mock_zaqar):
        mock_zc = mock.Mock()
        mock_zaqar.return_value = mock_zc
        message = mmod.Message('message', None, None, id=UUID)
        queue_name = 'senlin-receiver-%s' % message.id
        kwargs = {
            '_max_messages_post_size': 262144,
            '_default_message_ttl': 3600,
            'description': 'Queue for Senlin receiver.',
            'name': queue_name
        }
        mock_zc.queue_create.return_value = queue_name
        res = message._create_queue()

        self.assertEqual(queue_name, res)
        mock_zc.queue_create.assert_called_once_with(**kwargs)

    @mock.patch.object(mmod.Message, 'zaqar')
    def test__create_queue_fail(self, mock_zaqar):
        mock_zc = mock.Mock()
        mock_zaqar.return_value = mock_zc
        message = mmod.Message('message', None, None, id=UUID)
        queue_name = 'senlin-receiver-%s' % message.id
        kwargs = {
            '_max_messages_post_size': 262144,
            '_default_message_ttl': 3600,
            'description': 'Queue for Senlin receiver.',
            'name': queue_name
        }
        mock_zc.queue_create.side_effect = exception.InternalError()
        self.assertRaises(exception.EResourceCreation, message._create_queue)
        mock_zc.queue_create.assert_called_once_with(**kwargs)

    @mock.patch.object(mmod.Message, '_generate_subscriber_url')
    @mock.patch.object(mmod.Message, '_build_trust')
    @mock.patch.object(mmod.Message, 'zaqar')
    def test__create_subscription(self, mock_zaqar, mock_build_trust,
                                  mock_generate_subscriber_url):
        mock_zc = mock.Mock()
        mock_zaqar.return_value = mock_zc
        mock_build_trust.return_value = '123abc'
        subscriber = 'subscriber_url'
        mock_generate_subscriber_url.return_value = subscriber
        message = mmod.Message('message', None, None, id=UUID)
        queue_name = 'test-queue'
        kwargs = {
            "ttl": 3600,
            "subscriber": subscriber,
            "options": {
                "from": "senlin and zaqar",
                "subject": "hello, senlin",
                "trust_id": "123abc"
            }
        }
        mock_zc.subscription_create.return_value = 'subscription'
        res = message._create_subscription(queue_name)

        self.assertEqual('subscription', res)
        mock_generate_subscriber_url.assert_called_once_with()
        mock_zc.subscription_create.assert_called_once_with(queue_name,
                                                            **kwargs)

    @mock.patch.object(mmod.Message, '_generate_subscriber_url')
    @mock.patch.object(mmod.Message, '_build_trust')
    @mock.patch.object(mmod.Message, 'zaqar')
    def test__create_subscription_fail(self, mock_zaqar, mock_build_trust,
                                       mock_generate_subscriber_url):
        mock_zc = mock.Mock()
        mock_zaqar.return_value = mock_zc
        mock_build_trust.return_value = '123abc'
        subscriber = 'subscriber_url'
        mock_generate_subscriber_url.return_value = subscriber
        message = mmod.Message('message', None, None, id=UUID)
        queue_name = 'test-queue'
        kwargs = {
            "ttl": 3600,
            "subscriber": subscriber,
            "options": {
                "from": "senlin and zaqar",
                "subject": "hello, senlin",
                "trust_id": "123abc"
            }
        }

        mock_zc.subscription_create.side_effect = exception.InternalError()
        self.assertRaises(exception.EResourceCreation,
                          message._create_subscription, queue_name)
        mock_generate_subscriber_url.assert_called_once_with()
        mock_zc.subscription_create.assert_called_once_with(queue_name,
                                                            **kwargs)

    @mock.patch.object(mmod.Message, 'zaqar')
    def test_release_channel(self, mock_zaqar):
        mock_zc = mock.Mock()
        mock_zaqar.return_value = mock_zc
        channel = {'queue_name': 'test-queue',
                   'subscription': 'test-subscription-id'}
        message = mmod.Message('message', None, None, id=UUID,
                               channel=channel)

        message.release_channel()
        mock_zc.subscription_delete.assert_called_once_with(
            'test-queue', 'test-subscription-id')
        mock_zc.queue_delete.assert_called_once_with('test-queue')

    @mock.patch.object(mmod.Message, 'zaqar')
    def test_release_channel_subscription_delete_fail(self, mock_zaqar):
        mock_zc = mock.Mock()
        mock_zaqar.return_value = mock_zc
        channel = {'queue_name': 'test-queue',
                   'subscription': 'test-subscription-id'}
        message = mmod.Message('message', None, None, id=UUID,
                               channel=channel)
        mock_zc.subscription_delete.side_effect = exception.InternalError()

        self.assertRaises(exception.EResourceDeletion,
                          message.release_channel)
        mock_zc.subscription_delete.assert_called_once_with(
            'test-queue', 'test-subscription-id')

    @mock.patch.object(mmod.Message, 'zaqar')
    def test_release_channel_queue_delete_fail(self, mock_zaqar):
        mock_zc = mock.Mock()
        mock_zaqar.return_value = mock_zc
        channel = {'queue_name': 'test-queue',
                   'subscription': 'test-subscription-id'}
        message = mmod.Message('message', None, None, id=UUID,
                               channel=channel)
        mock_zc.queue_delete.side_effect = exception.InternalError()

        self.assertRaises(exception.EResourceDeletion,
                          message.release_channel)
        mock_zc.subscription_delete.assert_called_once_with(
            'test-queue', 'test-subscription-id')
        mock_zc.queue_delete.assert_called_once_with('test-queue')

    @mock.patch.object(ks_loading, 'load_auth_from_conf_options')
    @mock.patch.object(ks_loading, 'load_session_from_conf_options')
    @mock.patch.object(mmod.Message, 'keystone')
    def test__build_trust_exists(self, mock_keystone, mock_load_session,
                                 mock_load_auth):
        mock_auth = mock.Mock()
        mock_session = mock.Mock()
        mock_session.get_user_id.return_value = 'zaqar-trustee-user-id'
        mock_load_session.return_value = mock_session
        mock_load_auth.return_value = mock_auth
        mock_kc = mock.Mock()
        mock_keystone.return_value = mock_kc
        mock_trust = mock.Mock()
        mock_trust.id = 'mock-trust-id'
        message = mmod.Message('message', None, None, id=UUID,
                               user='user1', project='project1')
        mock_kc.trust_get_by_trustor.return_value = mock_trust

        res = message._build_trust()

        self.assertEqual('mock-trust-id', res)
        mock_kc.trust_get_by_trustor.assert_called_once_with(
            'user1', 'zaqar-trustee-user-id', 'project1')
        mock_load_auth.assert_called_once_with(cfg.CONF, 'zaqar')
        mock_load_session.assert_called_once_with(cfg.CONF, 'zaqar')
        mock_session.get_user_id.assert_called_once_with(auth=mock_auth)

    @mock.patch.object(ks_loading, 'load_auth_from_conf_options')
    @mock.patch.object(ks_loading, 'load_session_from_conf_options')
    @mock.patch.object(mmod.Message, 'keystone')
    def test__build_trust_create_new(self, mock_keystone, mock_load_session,
                                     mock_load_auth):
        mock_auth = mock.Mock()
        mock_session = mock.Mock()
        mock_session.get_user_id.return_value = 'zaqar-trustee-user-id'
        mock_load_session.return_value = mock_session
        mock_load_auth.return_value = mock_auth
        mock_kc = mock.Mock()
        mock_keystone.return_value = mock_kc
        mock_trust = mock.Mock()
        mock_trust.id = 'mock-trust-id'
        message = mmod.Message('message', None, None, id=UUID,
                               user='user1', project='project1')
        mock_kc.trust_get_by_trustor.return_value = None
        mock_kc.trust_create.return_value = mock_trust

        res = message._build_trust()

        self.assertEqual('mock-trust-id', res)
        mock_kc.trust_get_by_trustor.assert_called_once_with(
            'user1', 'zaqar-trustee-user-id', 'project1')
        mock_kc.trust_create.assert_called_once_with(
            'user1', 'zaqar-trustee-user-id', 'project1', ['admin'])

    @mock.patch.object(ks_loading, 'load_auth_from_conf_options')
    @mock.patch.object(ks_loading, 'load_session_from_conf_options')
    @mock.patch.object(mmod.Message, 'keystone')
    def test__build_trust_create_new_trust_failed(self, mock_keystone,
                                                  mock_load_session,
                                                  mock_load_auth):
        mock_auth = mock.Mock()
        mock_session = mock.Mock()
        mock_session.get_user_id.return_value = 'zaqar-trustee-user-id'
        mock_load_session.return_value = mock_session
        mock_load_auth.return_value = mock_auth
        mock_kc = mock.Mock()
        mock_keystone.return_value = mock_kc
        mock_trust = mock.Mock()
        mock_trust.id = 'mock-trust-id'
        message = mmod.Message('message', None, None, id=UUID,
                               user='user1', project='project1')
        mock_kc.trust_get_by_trustor.return_value = None
        mock_kc.trust_create.side_effect = exception.InternalError()

        self.assertRaises(exception.EResourceCreation,
                          message._build_trust)

        mock_kc.trust_get_by_trustor.assert_called_once_with(
            'user1', 'zaqar-trustee-user-id', 'project1')
        mock_kc.trust_create.assert_called_once_with(
            'user1', 'zaqar-trustee-user-id', 'project1', ['admin'])

    @mock.patch.object(ks_loading, 'load_auth_from_conf_options')
    @mock.patch.object(ks_loading, 'load_session_from_conf_options')
    @mock.patch.object(mmod.Message, 'keystone')
    def test__build_trust_get_trust_exception(self, mock_keystone,
                                              mock_load_session,
                                              mock_load_auth):
        mock_auth = mock.Mock()
        mock_session = mock.Mock()
        mock_session.get_user_id.return_value = 'zaqar-trustee-user-id'
        mock_load_session.return_value = mock_session
        mock_load_auth.return_value = mock_auth
        mock_kc = mock.Mock()
        mock_keystone.return_value = mock_kc
        mock_trust = mock.Mock()
        mock_trust.id = 'mock-trust-id'
        message = mmod.Message('message', None, None, id=UUID,
                               user='user1', project='project1')
        mock_kc.trust_get_by_trustor.side_effect = exception.InternalError()

        self.assertRaises(exception.EResourceCreation,
                          message._build_trust)

        mock_kc.trust_get_by_trustor.assert_called_once_with(
            'user1', 'zaqar-trustee-user-id', 'project1')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
from tempest.lib import decorators
from tempest.lib import exceptions
from tempest import test

from senlin.tests.tempest.api import base
from senlin.tests.tempest.common import constants
from senlin.tests.tempest.common import utils


class TestClusterUpdateNegativeNotFound(base.BaseSenlinAPITest):

    @test.attr(type=['negative'])
    @decorators.idempotent_id('f7a97fce-f495-44a8-b41a-7408139adacf')
    def test_cluster_update_cluster_not_found(self):
        self.assertRaises(exceptions.NotFound,
                          self.client.update_obj,
                          'clusters', 'f7a97fce-f495-44a8-b41a-7408139adacf',
                          {'cluster': {'name': 'new-name'}})


class TestClusterUpdateNegativeProfileNotFound(base.BaseSenlinAPITest):

    def setUp(self):
        super(TestClusterUpdateNegativeProfileNotFound, self).setUp()
        # Create a profile
        profile_id = utils.create_a_profile(self)
        self.addCleanup(utils.delete_a_profile, self, profile_id)
        self.profile_id = profile_id
        # Create a cluster
        self.cluster_id = utils.create_a_cluster(self, profile_id)
        self.addCleanup(utils.delete_a_cluster, self, self.cluster_id)

    @test.attr(type=['negative'])
    @decorators.idempotent_id('fb68921d-1fe8-4c14-be9a-51fa43d4f705')
    def test_cluster_update_profile_not_found(self):
        # Provided profile can not be found
        params = {
            'cluster': {
                'profile_id': 'fb68921d-1fe8-4c14-be9a-51fa43d4f705',
            }
        }
        # Verify badrequest exception(400) is raised.
        self.assertRaises(exceptions.BadRequest,
                          self.client.update_obj,
                          'clusters', self.cluster_id, params)


class TestClusterUpdateNegativeProfileMultichoices(base.BaseSenlinAPITest):
    def setUp(self):
        super(TestClusterUpdateNegativeProfileMultichoices, self).setUp()
        # Create a profile
        profile_id = utils.create_a_profile(self)
        self.addCleanup(utils.delete_a_profile, self, profile_id)
        self.profile_id = profile_id
        # Create a cluster
        self.cluster_id = utils.create_a_cluster(self, profile_id)
        self.addCleanup(utils.delete_a_cluster, self, self.cluster_id)
        # Create two new profiles of the same type with the same name
        new_spec = copy.deepcopy(constants.spec_nova_server)
        new_spec['properties']['flavor'] = 'new_flavor'
        new_profile_id1 = utils.create_a_profile(self, new_spec, name='p-nova')
        new_profile_id2 = utils.create_a_profile(self, new_spec, name='p-nova')
        self.addCleanup(utils.delete_a_profile, self, new_profile_id1)
        self.addCleanup(utils.delete_a_profile, self, new_profile_id2)

    @test.attr(type=['negative'])
    @decorators.idempotent_id('6a3eb86c-4b5c-4cfc-891c-7b0be17715f2')
    def test_cluster_update_profile_multichoices(self):
        # Multiple profiles are found for given name
        params = {
            'cluster': {
                'profile_id': 'p-nova',
            }
        }

        # Verify badrequest exception(400) is raised.
        self.assertRaises(exceptions.BadRequest,
                          self.client.update_obj,
                          'clusters', self.cluster_id, params)


class TestClusterUpdateNegativeProfileTypeUnmatch(base.BaseSenlinAPITest):
    def setUp(self):
        super(TestClusterUpdateNegativeProfileTypeUnmatch, self).setUp()
        # Create a profile
        profile_id = utils.create_a_profile(self)
        self.addCleanup(utils.delete_a_profile, self, profile_id)
        self.profile_id = profile_id
        # Create a cluster
        self.cluster_id = utils.create_a_cluster(self, profile_id)
        self.addCleanup(utils.delete_a_cluster, self, self.cluster_id)
        # Create a new profile of different type
        self.new_profile_id = utils.create_a_profile(
            self, spec=constants.spec_heat_stack)
        self.addCleanup(utils.delete_a_profile, self, self.new_profile_id)

    @test.attr(type=['negative'])
    @decorators.idempotent_id('c28e1bc0-fadb-4394-b2d0-67ad8b87ac04')
    def test_cluster_update_profile_type_unmatch(self):
        # New profile type is different from original cone
        params = {
            'cluster': {
                'profile_id': self.new_profile_id,
            }
        }

        # Verify badrequest exception(400) is raised.
        self.assertRaises(exceptions.BadRequest,
                          self.client.update_obj,
                          'clusters', self.cluster_id, params)


class TestClusterUpdateNegativeNoPropertyUpdated(base.BaseSenlinAPITest):

    def setUp(self):
        super(TestClusterUpdateNegativeNoPropertyUpdated, self).setUp()
        # Create a profile
        profile_id = utils.create_a_profile(self)
        self.addCleanup(utils.delete_a_profile, self, profile_id)
        self.profile_id = profile_id
        # Create a cluster
        self.cluster_id = utils.create_a_cluster(self, profile_id)
        self.addCleanup(utils.delete_a_cluster, self, self.cluster_id)

    @test.attr(type=['negative'])
    @decorators.idempotent_id('0fbd8fd9-7789-47da-b806-d91631a28556')
    def test_cluster_update_no_property_updated(self):
        # No any property is updated
        params = {
            'cluster': {}
        }
        # Verify badrequest exception(400) is raised.
        self.assertRaises(exceptions.BadRequest,
                          self.client.update_obj,
                          'clusters', self.cluster_id, params)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from senlin.common import consts
from senlin.common import scaleutils as su
from senlin.engine import health_manager
from senlin.objects import cluster as co
from senlin.policies import base as pb
from senlin.policies import health_policy
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class TestHealthPolicy(base.SenlinTestCase):

    def setUp(self):
        super(TestHealthPolicy, self).setUp()
        self.context = utils.dummy_context()

        self.spec = {
            'type': 'senlin.policy.health',
            'version': '1.0',
            'properties': {
                'detection': {
                    'type': 'NODE_STATUS_POLLING',
                    'options': {
                        'interval': 60
                    }
                },
                'recovery': {
                    'fencing': ['COMPUTE'],
                    'actions': ['REBUILD']
                }
            }
        }
        cluster = mock.Mock()
        cluster.id = 'CLUSTER_ID'
        node = mock.Mock()
        node.status = 'ACTIVE'
        cluster.nodes = [node]
        self.cluster = cluster
        self.patch('senlin.rpc.client.EngineClient')
        self.hp = health_policy.HealthPolicy('test-policy', self.spec)

    def test_policy_init(self):
        self.assertIsNone(self.hp.id)
        self.assertEqual('test-policy', self.hp.name)
        self.assertEqual('senlin.policy.health-1.0', self.hp.type)
        self.assertEqual('NODE_STATUS_POLLING', self.hp.check_type)
        self.assertEqual(60, self.hp.interval)
        self.assertEqual(['REBUILD'], self.hp.recover_actions)

    @mock.patch.object(health_manager, 'register')
    def test_attach(self, mock_hm_reg):

        policy_data = {
            'HealthPolicy': {
                'data': {
                    'check_type': self.hp.check_type,
                    'interval': self.hp.interval},
                'version': '1.0'
            }
        }

        res, data = self.hp.attach(self.cluster)
        self.assertTrue(res)
        self.assertEqual(policy_data, data)
        kwargs = {
            'check_type': self.hp.check_type,
            'interval': self.hp.interval,
            'params': {},
        }
        mock_hm_reg.assert_called_once_with('CLUSTER_ID',
                                            engine_id=None,
                                            **kwargs)

    @mock.patch.object(health_manager, 'unregister')
    def test_detach(self, mock_hm_reg):
        res, data = self.hp.detach(self.cluster)
        self.assertTrue(res)
        self.assertEqual('', data)
        mock_hm_reg.assert_called_once_with('CLUSTER_ID')

    def test_pre_op_default(self):
        action = mock.Mock(context='action_context', data={},
                           action=consts.CLUSTER_RECOVER)

        res = self.hp.pre_op(self.cluster.id, action)

        self.assertTrue(res)
        data = {
            'health': {
                'recover_action': ['REBUILD'],
                'fencing': ['COMPUTE'],
            }
        }
        self.assertEqual(data, action.data)

    @mock.patch.object(health_manager, 'disable')
    def test_pre_op_scale_in(self, mock_disable):
        action = mock.Mock(context='action_context', data={},
                           action=consts.CLUSTER_SCALE_IN)

        res = self.hp.pre_op(self.cluster.id, action)

        self.assertTrue(res)
        mock_disable.assert_called_once_with(self.cluster.id)

    @mock.patch.object(health_manager, 'disable')
    def test_pre_op_cluster_del_nodes(self, mock_disable):
        action = mock.Mock(context='action_context', data={},
                           action=consts.CLUSTER_DEL_NODES)

        res = self.hp.pre_op(self.cluster.id, action)

        self.assertTrue(res)
        mock_disable.assert_called_once_with(self.cluster.id)

    @mock.patch.object(health_manager, 'disable')
    def test_pre_op_node_delete(self, mock_disable):
        action = mock.Mock(context='action_context', data={},
                           action=consts.NODE_DELETE)

        res = self.hp.pre_op(self.cluster.id, action)

        self.assertTrue(res)
        mock_disable.assert_called_once_with(self.cluster.id)

    @mock.patch.object(health_manager, 'disable')
    def test_pre_op_resize_with_data(self, mock_disable):
        action = mock.Mock(context='action_context', data={'deletion': 'foo'},
                           action=consts.CLUSTER_RESIZE)

        res = self.hp.pre_op(self.cluster.id, action)

        self.assertTrue(res)
        mock_disable.assert_called_once_with(self.cluster.id)

    @mock.patch.object(su, 'parse_resize_params')
    @mock.patch.object(co.Cluster, 'get')
    @mock.patch.object(health_manager, 'disable')
    def test_pre_op_resize_without_data(self, mock_disable, mock_get,
                                        mock_parse):
        def fake_check(*args, **kwargs):
            action.data['deletion'] = {'foo': 'bar'}
            return pb.CHECK_OK, 'good'

        x_cluster = mock.Mock()
        mock_get.return_value = x_cluster
        action = mock.Mock(context='action_context', data={},
                           action=consts.CLUSTER_RESIZE)
        mock_parse.side_effect = fake_check

        res = self.hp.pre_op(self.cluster.id, action)

        self.assertTrue(res)
        mock_disable.assert_called_once_with(self.cluster.id)

    @mock.patch.object(su, 'parse_resize_params')
    @mock.patch.object(co.Cluster, 'get')
    @mock.patch.object(health_manager, 'disable')
    def test_pre_op_resize_parse_error(self, mock_disable, mock_get,
                                       mock_parse):
        x_cluster = mock.Mock()
        mock_get.return_value = x_cluster
        action = mock.Mock(context='action_context', data={},
                           action=consts.CLUSTER_RESIZE)
        mock_parse.return_value = pb.CHECK_ERROR, 'no good'

        res = self.hp.pre_op(self.cluster.id, action)

        self.assertFalse(res)
        self.assertEqual(pb.CHECK_ERROR, action.data['status'])
        self.assertEqual('no good', action.data['reason'])
        self.assertEqual(0, mock_disable.call_count)

    def test_post_op_default(self):
        action = mock.Mock(action='FAKE_ACTION')

        res = self.hp.post_op(self.cluster.id, action)

        self.assertTrue(res)

    @mock.patch.object(health_manager, 'enable')
    def test_post_op_scale_in(self, mock_enable):
        action = mock.Mock(action=consts.CLUSTER_SCALE_IN)

        res = self.hp.post_op(self.cluster.id, action)

        self.assertTrue(res)
        mock_enable.assert_called_once_with(self.cluster.id)

    @mock.patch.object(health_manager, 'enable')
    def test_post_op_cluster_del_nodes(self, mock_enable):
        action = mock.Mock(action=consts.CLUSTER_DEL_NODES)

        res = self.hp.post_op(self.cluster.id, action)

        self.assertTrue(res)
        mock_enable.assert_called_once_with(self.cluster.id)

    @mock.patch.object(health_manager, 'enable')
    def test_post_op_node_delete(self, mock_enable):
        action = mock.Mock(action=consts.NODE_DELETE)

        res = self.hp.post_op(self.cluster.id, action)

        self.assertTrue(res)
        mock_enable.assert_called_once_with(self.cluster.id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import eventlet

from senlin.common.i18n import _
from senlin.common import scaleutils
from senlin.engine.actions import base
from senlin.engine import cluster as cm
from senlin.engine import event as EVENT
from senlin.engine import node as node_mod
from senlin.engine import senlin_lock
from senlin.policies import base as pb


class NodeAction(base.Action):
    """An action that can be performed on a cluster member (node)."""

    ACTIONS = (
        NODE_CREATE, NODE_DELETE, NODE_UPDATE,
        NODE_JOIN, NODE_LEAVE,
        NODE_CHECK, NODE_RECOVER
    ) = (
        'NODE_CREATE', 'NODE_DELETE', 'NODE_UPDATE',
        'NODE_JOIN', 'NODE_LEAVE',
        'NODE_CHECK', 'NODE_RECOVER'
    )

    def __init__(self, target, action, context, **kwargs):
        """Constructor for a node action object.

        :param target: ID of the target node object on which the action is to
                       be executed.
        :param action: The name of the action to be executed.
        :param context: The context used for accessing the DB layer.
        :param dict kwargs: Additional parameters that can be passed to the
                            action.
        """
        super(NodeAction, self).__init__(target, action, context, **kwargs)

        try:
            self.node = node_mod.Node.load(self.context, node_id=self.target)
        except Exception:
            self.node = None

    def do_create(self):
        """Handler for the NODE_CREATE action.

        :returns: A tuple containing the result and the corresponding reason.
        """
        if self.node.cluster_id and self.cause == base.CAUSE_RPC:
            # If node is created with target cluster specified,
            # check cluster size constraint
            cluster = cm.Cluster.load(self.context, self.node.cluster_id)
            result = scaleutils.check_size_params(
                cluster, cluster.desired_capacity + 1, None, None, True)

            if result:
                return self.RES_ERROR, result
            # Update cluster desired_capacity if node is already in db.
            cluster.desired_capacity += 1
            cluster.add_node(self.node)
            cluster.store(self.context)

        res = self.node.do_create(self.context)
        if res:
            return self.RES_OK, _('Node created successfully.')
        else:
            return self.RES_ERROR, _('Node creation failed.')

    def do_delete(self):
        """Handler for the NODE_DELETE action.

        :returns: A tuple containing the result and the corresponding reason.
        """
        if self.node.cluster_id and self.cause == base.CAUSE_RPC:
            # If node belongs to a cluster, check size constraint
            # before deleting it
            cluster = cm.Cluster.load(self.context, self.node.cluster_id)
            result = scaleutils.check_size_params(cluster,
                                                  cluster.desired_capacity - 1,
                                                  None, None, True)
            if result:
                return self.RES_ERROR, result

            # handle grace_period
            pd = self.data.get('deletion', None)
            if pd:
                grace_period = pd.get('grace_period', 0)
                if grace_period:
                    eventlet.sleep(grace_period)
            # check if desired_capacity should be changed
            do_reduce = True
            pd = self.data.get('deletion', None)
            if pd:
                do_reduce = pd.get('reduce_desired_capacity', True)
            if do_reduce:
                cluster.desired_capacity -= 1
                cluster.store(self.context)
            cluster.remove_node(self.node.id)

        res = self.node.do_delete(self.context)
        if res:
            return self.RES_OK, _('Node deleted successfully.')
        else:
            return self.RES_ERROR, _('Node deletion failed.')

    def do_update(self):
        """Handler for the NODE_UPDATE action.

        :returns: A tuple containing the result and the corresponding reason.
        """
        params = self.inputs
        res = self.node.do_update(self.context, params)
        if res:
            return self.RES_OK, _('Node updated successfully.')
        else:
            return self.RES_ERROR, _('Node update failed.')

    def do_join(self):
        """Handler for the NODE_JOIN action.

        :returns: A tuple containing the result and the corresponding reason.
        """
        cluster_id = self.inputs.get('cluster_id')
        # Check the size constraint of parent cluster
        cluster = cm.Cluster.load(self.context, cluster_id)
        new_capacity = cluster.desired_capacity + 1
        result = scaleutils.check_size_params(cluster, new_capacity,
                                              None, None, True)
        if result:
            return self.RES_ERROR, result

        result = self.node.do_join(self.context, cluster_id)
        if result:
            cluster.add_node(self.node)
            return self.RES_OK, _('Node successfully joined cluster.')
        else:
            return self.RES_ERROR, _('Node failed in joining cluster.')

    def do_leave(self):
        """Handler for the NODE_LEAVE action.

        :returns: A tuple containing the result and the corresponding reason.
        """
        # Check the size constraint of parent cluster
        cluster = cm.Cluster.load(self.context, self.node.cluster_id)
        new_capacity = cluster.desired_capacity - 1
        result = scaleutils.check_size_params(cluster, new_capacity,
                                              None, None, True)
        if result:
            return self.RES_ERROR, result

        res = self.node.do_leave(self.context)
        if res:
            cluster.remove_node(self.node.id)
            return self.RES_OK, _('Node successfully left cluster.')
        else:
            return self.RES_ERROR, _('Node failed in leaving cluster.')

    def do_check(self):
        """Handler for the NODE_check action.

        :returns: A tuple containing the result and the corresponding reason.
        """
        res = self.node.do_check(self.context)
        if res:
            return self.RES_OK, _('Node status is ACTIVE.')
        else:
            return self.RES_ERROR, _('Node status is not ACTIVE.')

    def do_recover(self):
        """Handler for the NODE_RECOVER action.

        :returns: A tuple containing the result and the corresponding reason.
        """
        res = self.node.do_recover(self.context, **self.inputs)
        if res:
            return self.RES_OK, _('Node recovered successfully.')
        else:
            return self.RES_ERROR, _('Node recover failed.')

    def _execute(self):
        """Private function that finds out the handler and execute it."""

        action_name = self.action.lower()
        method_name = action_name.replace('node', 'do')
        method = getattr(self, method_name, None)

        if method is None:
            reason = _('Unsupported action: %s') % self.action
            EVENT.error(self.context, self.node, self.action, 'Failed', reason)
            return self.RES_ERROR, reason

        return method()

    def execute(self, **kwargs):
        """Interface function for action execution.

        :param dict kwargs: Parameters provided to the action, if any.
        :returns: A tuple containing the result and the related reason.
        """
        # Since node.cluster_id could be reset to '' during action execution,
        # we record it here for policy check and cluster lock release.
        saved_cluster_id = self.node.cluster_id
        if (saved_cluster_id and self.cause == base.CAUSE_RPC):
            res = senlin_lock.cluster_lock_acquire(
                self.context, self.node.cluster_id, self.id, self.owner,
                senlin_lock.NODE_SCOPE, False)

            if not res:
                return self.RES_RETRY, _('Failed in locking cluster')

            self.policy_check(self.node.cluster_id, 'BEFORE')
            if self.data['status'] != pb.CHECK_OK:
                # Don't emit message since policy_check should have done it
                senlin_lock.cluster_lock_release(saved_cluster_id, self.id,
                                                 senlin_lock.NODE_SCOPE)
                return self.RES_ERROR, 'Policy check: ' + self.data['reason']

        reason = ''
        try:
            res = senlin_lock.node_lock_acquire(self.context, self.node.id,
                                                self.id, self.owner, False)
            if not res:
                res = self.RES_ERROR
                reason = _('Failed in locking node')
            else:
                res, reason = self._execute()
                if (res == self.RES_OK and saved_cluster_id and
                        self.cause == base.CAUSE_RPC):
                    self.policy_check(saved_cluster_id, 'AFTER')
                    if self.data['status'] != pb.CHECK_OK:
                        res = self.RES_ERROR
                        reason = 'Policy check: ' + self.data['reason']
                    else:
                        res = self.RES_OK
        finally:
            senlin_lock.node_lock_release(self.node.id, self.id)
            if saved_cluster_id and self.cause == base.CAUSE_RPC:
                senlin_lock.cluster_lock_release(saved_cluster_id, self.id,
                                                 senlin_lock.NODE_SCOPE)
        return res, reason

    def cancel(self):
        """Handler for cancelling the action."""
        return self.RES_OK



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from oslo_serialization import jsonutils

from senlin.api.openstack.v1 import webhooks
from senlin.common import policy
from senlin.rpc import client as rpc_client
from senlin.tests.unit.api import shared
from senlin.tests.unit.common import base


@mock.patch.object(policy, 'enforce')
class WebhookControllerTest(shared.ControllerTest, base.SenlinTestCase):
    def setUp(self):
        super(WebhookControllerTest, self).setUp()

        class DummyConfig(object):
            bind_port = 8778

        cfgopts = DummyConfig()
        self.controller = webhooks.WebhookController(options=cfgopts)

    def test_webhook_trigger(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'trigger', True)
        body = None
        webhook_id = 'test_webhook_id'
        action_id = 'test_action_id'

        engine_response = {
            'action': action_id,
        }

        req = self._post('/webhooks/test_webhook_id/trigger',
                         jsonutils.dumps(body))
        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=engine_response)

        resp = self.controller.trigger(req, webhook_id=webhook_id, body=None)

        mock_call.assert_called_with(
            req.context,
            ('webhook_trigger', {
                'params': None,
                'identity': webhook_id,
            })
        )

        expected = engine_response
        location = {'location': '/actions/test_action_id'}
        expected.update(location)
        self.assertEqual(expected, resp)

    def test_webhook_trigger_with_params(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'trigger', True)
        body = {'params': {'key': 'value'}}
        webhook_id = 'test_webhook_id'

        engine_response = {'action': 'FAKE_ACTION'}

        req = self._post('/webhooks/test_webhook_id/trigger',
                         jsonutils.dumps(body))
        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=engine_response)

        resp = self.controller.trigger(req, webhook_id=webhook_id, body=body)

        mock_call.assert_called_with(req.context,
                                     ('webhook_trigger',
                                      {'params': {'key': 'value'},
                                       'identity': webhook_id}))

        expected = engine_response
        self.assertEqual(expected, resp)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from senlin.objects import base as obj_base
from senlin.objects import fields as obj_fields
from senlin.tests.unit.common import base


class TestBaseObject(base.SenlinTestCase):

    def test_base_class(self):
        obj = obj_base.SenlinObject()
        self.assertEqual(obj_base.SenlinObject.OBJ_PROJECT_NAMESPACE,
                         obj.OBJ_PROJECT_NAMESPACE)
        self.assertEqual(obj_base.SenlinObject.VERSION,
                         obj.VERSION)

    @mock.patch.object(obj_base.SenlinObject, "obj_reset_changes")
    def test_from_db_object(self, mock_obj_reset_ch):
        class TestSenlinObject(obj_base.SenlinObject,
                               obj_base.VersionedObjectDictCompat):
            fields = {
                "key1": obj_fields.StringField(),
                "key2": obj_fields.StringField(),
                "metadata": obj_fields.JsonField()
            }

        obj = TestSenlinObject()
        context = mock.Mock()
        db_obj = {
            "key1": "value1",
            "key2": "value2",
            "meta_data": {"key3": "value3"}
        }
        res = obj_base.SenlinObject._from_db_object(context, obj, db_obj)
        self.assertIsNotNone(res)
        self.assertEqual("value1", obj["key1"])
        self.assertEqual("value2", obj["key2"])
        self.assertEqual({"key3": "value3"}, obj["metadata"])
        self.assertEqual(obj._context, context)
        mock_obj_reset_ch.assert_called_once_with()

    def test_from_db_object_none(self):
        obj = obj_base.SenlinObject()
        db_obj = None
        context = mock.Mock()

        res = obj_base.SenlinObject._from_db_object(context, obj, db_obj)
        self.assertIsNone(res)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Policy for scheduling nodes across availability zones.

NOTE: For full documentation about how the policy works, check:
http://docs.openstack.org/developer/senlin/developer/policies/zone_v1.html
"""

import math
import operator

from oslo_log import log as logging

from senlin.common import consts
from senlin.common import exception as exc
from senlin.common.i18n import _, _LE
from senlin.common import scaleutils
from senlin.common import schema
from senlin.engine import cluster as cm
from senlin.objects import cluster as co
from senlin.policies import base

LOG = logging.getLogger(__name__)


class ZonePlacementPolicy(base.Policy):
    """Policy for placing members of a cluster across availability zones."""

    VERSION = '1.0'

    PRIORITY = 300

    TARGET = [
        ('BEFORE', consts.CLUSTER_SCALE_OUT),
        ('BEFORE', consts.CLUSTER_SCALE_IN),
        ('BEFORE', consts.CLUSTER_RESIZE),
        ('BEFORE', consts.NODE_CREATE),
    ]

    PROFILE_TYPE = [
        'os.nova.server-1.0',
    ]

    KEYS = (
        ZONES,
    ) = (
        'zones',
    )

    _AZ_KEYS = (
        ZONE_NAME, ZONE_WEIGHT,
    ) = (
        'name', 'weight',
    )

    properties_schema = {
        ZONES: schema.List(
            _('List of availability zones to choose from.'),
            schema=schema.Map(
                _('An availability zone as candidate.'),
                schema={
                    ZONE_NAME: schema.String(
                        _('Name of an availability zone.'),
                    ),
                    ZONE_WEIGHT: schema.Integer(
                        _('Weight of the availability zone (default is 100).'),
                        default=100,
                        required=False,
                    )
                },
            ),
        ),
    }

    def __init__(self, name, spec, **kwargs):
        super(ZonePlacementPolicy, self).__init__(name, spec, **kwargs)

        self.zones = dict((z[self.ZONE_NAME], z[self.ZONE_WEIGHT])
                          for z in self.properties.get(self.ZONES))

    def validate(self, context, validate_props=False):
        super(ZonePlacementPolicy, self).validate(context, validate_props)

        if not validate_props:
            return True

        nc = self.nova(context.user, context.project)
        input_azs = sorted(self.zones.keys())
        valid_azs = nc.validate_azs(input_azs)
        invalid_azs = sorted(set(input_azs) - set(valid_azs))
        if invalid_azs:
            msg = _("The specified %(key)s '%(value)s' could not be "
                    "found.") % {'key': self.ZONE_NAME,
                                 'value': list(invalid_azs)}
            raise exc.InvalidSpec(message=msg)

        return True

    def _create_plan(self, current, zones, count, expand):
        """Compute a placement plan based on the weights of AZs.

        :param current: Distribution of existing nodes.
        :returns: A dict that contains a placement plan.
        """
        # sort candidate zones by distribution and covert it into a list
        candidates = sorted(zones.items(), key=operator.itemgetter(1),
                            reverse=expand)

        sum_weight = sum(zones.values())
        if expand:
            total = count + sum(current.values())
        else:
            total = sum(current.values()) - count

        remain = count
        plan = dict.fromkeys(zones.keys(), 0)

        for i in range(len(zones)):
            zone = candidates[i][0]
            weight = candidates[i][1]
            q = total * weight / float(sum_weight)
            if expand:
                quota = int(math.ceil(q))
                headroom = quota - current[zone]
            else:
                quota = int(math.floor(q))
                headroom = current[zone] - quota

            if headroom <= 0:
                continue

            if headroom < remain:
                plan[zone] = headroom
                remain -= headroom
            else:
                plan[zone] = remain if remain > 0 else 0
                remain = 0
                break

        if remain > 0:
            return None

        # filter out zero values
        result = {}
        for z, c in plan.items():
            if c > 0:
                result[z] = c

        return result

    def _get_count(self, cluster_id, action):
        """Get number of nodes to create or delete.

        :param cluster_id: The ID of the target cluster.
        :param action: The action object which triggered this policy check.
        :return: An integer value which can be 1) positive - number of nodes
                 to create; 2) negative - number of nodes to delete; 3) 0 -
                 something wrong happened, and the policy check failed.
        """
        if action.action == consts.NODE_CREATE:
            # skip the policy if availability zone is specified in profile
            profile = action.node.rt['profile']
            if profile.properties[profile.AVAILABILITY_ZONE]:
                return 0
            return 1

        if action.action == consts.CLUSTER_RESIZE:
            if action.data.get('deletion', None):
                return -action.data['deletion']['count']
            elif action.data.get('creation', None):
                return action.data['creation']['count']

            db_cluster = co.Cluster.get(action.context, cluster_id)
            res = scaleutils.parse_resize_params(action, db_cluster)
            if res[0] == base.CHECK_ERROR:
                action.data['status'] = base.CHECK_ERROR
                action.data['reason'] = res[1]
                LOG.error(res[1])
                return 0

            if action.data.get('deletion', None):
                return -action.data['deletion']['count']
            else:
                return action.data['creation']['count']

        if action.action == consts.CLUSTER_SCALE_IN:
            pd = action.data.get('deletion', None)
            if pd is None:
                return -action.inputs.get('count', 1)
            else:
                return -pd.get('count', 1)

        # CLUSTER_SCALE_OUT: an action that inflates the cluster
        pd = action.data.get('creation', None)
        if pd is None:
            return action.inputs.get('count', 1)
        else:
            return pd.get('count', 1)

    def pre_op(self, cluster_id, action):
        """Callback function when cluster membership is about to change.

        :param cluster_id: ID of the target cluster.
        :param action: The action that triggers this policy check.
        """
        count = self._get_count(cluster_id, action)
        if count == 0:
            return

        expand = True
        if count < 0:
            expand = False
            count = -count

        cluster = cm.Cluster.load(action.context, cluster_id)

        nc = self.nova(cluster.user, cluster.project)
        zones_good = nc.validate_azs(self.zones.keys())
        if len(zones_good) == 0:
            action.data['status'] = base.CHECK_ERROR
            action.data['reason'] = _('No availability zone found available.')
            LOG.error(_LE('No availability zone found available.'))
            return

        zones = {}
        for z, w in self.zones.items():
            if z in zones_good:
                zones[z] = w

        current = cluster.get_zone_distribution(action.context, zones.keys())
        result = self._create_plan(current, zones, count, expand)

        if not result:
            action.data['status'] = base.CHECK_ERROR
            action.data['reason'] = _('There is no feasible plan to '
                                      'handle all nodes.')
            LOG.error(_LE('There is no feasible plan to handle all nodes.'))
            return

        if expand:
            if 'creation' not in action.data:
                action.data['creation'] = {}
            action.data['creation']['count'] = count
            action.data['creation']['zones'] = result
        else:
            if 'deletion' not in action.data:
                action.data['deletion'] = {}
            action.data['deletion']['count'] = count
            action.data['deletion']['zones'] = result
